/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.engine.orchestra.incoming;

import java.net.URI;

import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jms.MapMessage;
import javax.jms.MessageListener;
import javax.jms.Queue;
import javax.jms.QueueConnection;
import javax.jms.QueueConnectionFactory;
import javax.jms.QueueSender;
import javax.jms.QueueSession;
import javax.jms.Session;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.xml.namespace.QName;
import javax.xml.transform.dom.DOMSource;

import org.objectweb.orchestra.util.Util;
import org.objectweb.petals.component.common.bc.AbstractBindingComponent;
import org.objectweb.petals.component.common.util.MessageExchangeWrapper;
import org.objectweb.petals.component.common.util.MessageExchangeWrapperImpl;
import org.objectweb.petals.engine.orchestra.util.OrchestraUtil;
import org.objectweb.petals.engine.orchestra.util.SourceTransformer;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * Implementation of the JMS {@link MessageListener}. This listener will be
 * used by the JBI container incoming message listener. The binding component
 * will handle incoming JMS message and forward them to a JBI Endpoint.
 * 
 * @author Christophe HAMERLING - eBMWebSourcing
 * 
 */
public class OrchestraJMSMessageListener implements MessageListener {

    protected SourceTransformer transformer = null;

    protected DeliveryChannel channel = null;

    protected AbstractBindingComponent component = null;

    /**
     * Creates a new instance of {@link OrchestraJMSMessageListener}
     * 
     * @param component
     * @param jmsFactory
     * @param jmsSession
     * @param threadPool
     * @throws Exception
     */
    public OrchestraJMSMessageListener(AbstractBindingComponent component)
        throws MessagingException {
        this.component = component;
        this.channel = component.getContext().getDeliveryChannel();
        transformer = new SourceTransformer();
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jms.MessageListener#onMessage(javax.jms.Message)
     */
    public void onMessage(javax.jms.Message message) {
        Element newBody = null;
        MapMessage m = (MapMessage) message;
        String mepEnd = "in-out";

        try {
            String operationNS = m.getString("operationNS");
            String operationName = m.getString("operationName");
            String serviceNS = m.getString("serviceNS");
            String serviceName = m.getString("serviceName");
            String portTypeNS = m.getString("portTypeNS");
            String portTypeName = m.getString("portTypeName");
            String portName = m.getString("portName");
            String id = m.getString("id");
            boolean oneWay = m.getBoolean("oneWay");

            Element body = (Element) Util
                .getObjectFromBytes(m.getBytes("body"));

            if (oneWay) {
                mepEnd = "in-only";
            }

            URI mep = URI.create("http://www.w3.org/2004/08/wsdl/" + mepEnd);

            MessageExchangeFactory factory = channel.createExchangeFactory();
            MessageExchangeWrapper me = new MessageExchangeWrapperImpl(factory
                .createExchange(mep));

            me.setInterfaceName(new QName(portTypeNS, portTypeName));
            QName serviceQN = new QName(serviceNS, serviceName);
            me.setService(serviceQN);
            me.setEndpoint(this.component.getContext().getEndpoint(serviceQN,
                portName));
            me.setOperation(new QName(operationNS, operationName));
            NormalizedMessage nm = me.createMessage();
            me.setInMessage(nm);
            nm.setContent(new DOMSource(body));

            // send message exchange. Wait for response if the message is an
            // inout one
            if (oneWay) {
                this.channel.send(me);
            } else {
                boolean res = this.channel.sendSync(me);
                if (!res) {
                    throw new Exception(
                        "OrchestraMessageListener.onMessage, Timeout on sending request/response message");
                }
            }
            
            if (me.getStatus() == ExchangeStatus.ACTIVE) {
                if (me.getFault() != null) {
                    Document fault = null;
                    try {
                        fault = (Document) this.transformer.toDOMNode(me
                            .getFault());
                        me.setStatus(ExchangeStatus.DONE);
                    } catch (Exception e) {
                        me.setError(e);
                        throw new Exception(e);
                    } finally {
                        this.channel.send(me);
                    }
                    // TODO : if there is a fault, send it to perform the fault
                    // mechanism in Orchestra
                } else {
                    try {
                        NormalizedMessage out = me.getMessage("out");
                        if (out != null) {
                            Node tmp = this.transformer.toDOMNode(out);
                            if (tmp.getOwnerDocument() != null) {
                                newBody = tmp.getOwnerDocument()
                                    .getDocumentElement();
                            } else if (tmp instanceof Document) {
                                newBody = ((Document) tmp).getDocumentElement();
                            } else {
                                throw new Exception(
                                    "OrchestraMessageExchangeListener, out normalize message content is not instance of Document and it has not ownerDocument ! Its class is : "
                                        + tmp.getClass());
                            }
                            newBody = OrchestraUtil.removeEnvelope(newBody);
                        }
                    } catch (Exception e) {
                        throw new Exception(e);
                    }
                }
            } else if (me.getStatus() == ExchangeStatus.ERROR) {
                // Extract error
                Exception error = me.getError();
                throw new Exception(
                    "OrchestraMessageListener.onMessage, Unknown", error);
            }

            if (!oneWay && me.getStatus() == ExchangeStatus.ACTIVE) {
                if (newBody == null) {
                    throw new Exception(
                        "OrchestraMessageListener.onMessage, out message is empty");
                }
                Context jndi = new InitialContext();
                QueueConnectionFactory QCF = (QueueConnectionFactory) jndi
                    .lookup("QCF");
                QueueConnection connection = QCF.createQueueConnection();
                Session session = connection.createQueueSession(false,
                    Session.AUTO_ACKNOWLEDGE);
                QueueSender prod = null;
                MapMessage messageReturn = null;
                // sendMessage to OrchestraJbiSEQueue
                Queue OrchestraJbiSEQueue = (Queue) jndi
                    .lookup("OrchestraJbiSEQueue");
                prod = ((QueueSession) session)
                    .createSender(OrchestraJbiSEQueue);
                messageReturn = session.createMapMessage();
                messageReturn.setStringProperty("destination",
                    "OrchestraEngine");
                messageReturn.setStringProperty("id", id);
                byte[] byteTab = Util.getBytesFromObject(Util
                    .getNodeAsString(newBody));
                messageReturn.setBytes("newBody", byteTab);
                connection.start();
                prod.send(messageReturn);

                prod.close();
                session.close();
                connection.close();
                prod = null;
                session = null;
                connection = null;
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
}
