/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.engine.orchestra.outgoing;

import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.messaging.InOnly;
import javax.jbi.messaging.InOptionalOut;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.messaging.RobustInOnly;
import javax.xml.soap.MessageFactory;
import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPBodyElement;
import javax.xml.soap.SOAPConnection;
import javax.xml.soap.SOAPConnectionFactory;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPMessage;
import javax.xml.soap.SOAPPart;
import javax.xml.transform.dom.DOMSource;

import org.objectweb.petals.component.common.HandlingException;
import org.objectweb.petals.component.common.bc.JBIListener;
import org.objectweb.petals.component.common.util.MessageExchangeWrapper;
import org.objectweb.petals.component.common.util.NormalizedMessageUtil;
import org.objectweb.petals.component.common.util.PetalsExtensionsUtil;
import org.objectweb.petals.engine.orchestra.util.SourceTransformer;
import org.objectweb.petals.tools.jbicommon.descriptor.Extensions;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

/**
 * @author Christophe HAMERLING - eBMWebSourcing
 *
 */
public class OrchestraJBIListener implements JBIListener {
    
    /**
     * The logger
     */
    private Logger logger;
    
    /**
     * 
     * Creates a new instance of {@link OrchestraJBIListener}
     *
     */
    public OrchestraJBIListener(Logger logger) {
        this.logger = logger;
    }

    /* (non-Javadoc)
     * @see org.objectweb.petals.component.common.bc.JBIListener#onJBIMessage(java.lang.String, org.objectweb.petals.component.common.util.MessageExchangeWrapper, org.objectweb.petals.tools.jbicommon.descriptor.Extensions)
     */
    public boolean onJBIMessage(String address, MessageExchangeWrapper exchange, Extensions extensions) throws HandlingException {
        boolean result = false;
        String orchestraPointURL = null;

        NormalizedMessage inNM = exchange.getInMessage();
        String operation = exchange.getOperationName();
        Node response = null;
        
        try {
            // get orchestra external address
            orchestraPointURL = PetalsExtensionsUtil
                .extractValueFromKeyValueExtension(extensions,
                    PetalsExtensionsUtil.ADDRESS);

            // call Bsoap engine
            response = callOrchestraEngine(orchestraPointURL, operation, inNM);

            // process MEP
            if (exchange instanceof InOnly || exchange instanceof RobustInOnly) {
                if (response != null) {
                    throw new UnsupportedOperationException(
                        "Did not expect return value for in-only or robust-in-only");
                }

            } else if (exchange instanceof InOptionalOut) {
                if (response != null) {
                    fillOutMessage(exchange, response);
                    result = true;
                }

            } else if (exchange instanceof InOut) {
                if (response == null) {
                    throw new UnsupportedOperationException(
                        "Expected return data for in-out");
                } else {
                    fillOutMessage(exchange, response);
                    result = true;
                }

            } else {
                throw new UnsupportedOperationException("Unhandled mep: "
                    + exchange.getPattern());
            }
            
        } catch (Exception e) {
            throw new HandlingException(e);
        }
        return result;
    }

    /**
     * Send message exchange response to consumer. The properties and
     * attachements are copied from input NM to output NM. The response
     * {@link Node} is set has output content.
     * 
     * @param exchange
     * @param response
     * @throws Exception
     */
    protected void fillOutMessage(MessageExchangeWrapper exchange, Node response)
        throws MessagingException {
        DOMSource ds = new DOMSource(response);
        NormalizedMessage out = exchange.createMessage();
        out.setContent(ds);

        NormalizedMessage in = exchange.getInMessage();
        NormalizedMessageUtil.copyProperties(in, out);
        NormalizedMessageUtil.copyAttachments(in, out);
        exchange.setOutMessage(out);
    }

    /**
     * Call the orchestra engine. The engine is called with a SOAP message
     * {@link SOAPMessage}.
     * 
     * @param bsoapEndpointURL
     * @param operation
     *            TODO
     * @param in
     * 
     * @return
     * @throws Exception
     */
    protected Node callOrchestraEngine(String bsoapEndpointURL, String operation,
        NormalizedMessage inNM) throws Exception {

        logger.log(Level.FINE, "Calling BSOAP, endpoint = " + bsoapEndpointURL
            + " / operation = " + operation);

        // transform normalized message content
        SourceTransformer transformer = new SourceTransformer();
        Document inDoc = (Document) transformer.toDOMNode(inNM);

        // create SOAP connection
        SOAPConnectionFactory conFactory = SOAPConnectionFactory.newInstance();
        SOAPConnection connection = conFactory.createConnection();

        // create SOAP message
        MessageFactory messFactory = MessageFactory.newInstance();
        SOAPMessage message = messFactory.createMessage();
        SOAPPart soapPart = message.getSOAPPart();
        SOAPEnvelope requestEnvelope = soapPart.getEnvelope();
        SOAPBody soapBody = requestEnvelope.getBody();
        // create operation
        soapBody.addBodyElement(requestEnvelope
            .createName(operation));
        soapBody.addDocument(inDoc);

        // call bsoap engine with soap
        SOAPMessage response = connection.call(message, bsoapEndpointURL);
        connection.close();

        Node result = (SOAPBodyElement) response.getSOAPBody()
            .getChildElements().next();
        logger.log(Level.INFO, result.toString());

        return result;
    }


    
}
