/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.engine.orchestra.su;

import java.io.File;
import java.io.FilenameFilter;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLClassLoader;
import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.management.DeploymentException;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.spi.InitialContextFactory;

import org.objectweb.orchestra.admin.BpelAdminSession;
import org.objectweb.orchestra.exception.BPELAdminException;
import org.objectweb.petals.component.common.PEtALSComponentSDKException;
import org.objectweb.petals.component.common.su.ServiceUnitListener;
import org.objectweb.petals.component.common.util.PetalsExtensionsUtil;
import org.objectweb.petals.engine.orchestra.exception.OrchestraJbiException;
import org.objectweb.petals.engine.orchestra.incoming.JMSConnection;
import org.objectweb.petals.engine.orchestra.util.OrchestraConnection;
import org.objectweb.petals.tools.jbicommon.descriptor.Extensions;
import org.objectweb.petals.tools.jbicommon.descriptor.JBIDescriptor;

import static org.objectweb.petals.component.common.util.ClassLoaderUtil.createClassLoader;

/**
 * Used to deploy/undeploy the BPEL sheet into Orchestra
 * 
 * @author Christophe HAMERLING - eBMWebSourcing
 * 
 */
public class OrchestraServiceUnitListener implements ServiceUnitListener {

    /**
     * The logger
     */
    private Logger logger;

    /**
     * The file filter used to get only BPEL and WSDL files
     */
    private FilenameFilter filter;

    /**
     * List of deployed BPEl process names
     */
    private List<String> deployedProcesses;

    /**
     * 
     */
    private Map<String, String> jmsAddresses;

    /**
     * Consumers connections map
     */
    private Map<String, JMSConnection> jmsConsumerConnections;

    private static final String CONNECTION_FACTORY = "connection-factory";

    private static final String ADDRESS = "address";

    private static final String INITIAL_CONTEXT_FACTORY = "initial-context-factory";

    private static final String USER = "user";

    private static final String PASSWORD = "password";

    private static final String TRANSACTED = "transacted";

    /**
     * Creates a new instance of {@link OrchestraServiceUnitListener}
     * 
     * @param logger
     */
    public OrchestraServiceUnitListener(
        Map<String, JMSConnection> jmsConsumerConnections, Logger logger) {
        this.logger = logger;
        this.filter = new BpelFilter();
        this.jmsAddresses = new HashMap<String, String>();
        this.jmsConsumerConnections = jmsConsumerConnections;
        this.deployedProcesses = new Vector<String>();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.component.common.ServiceUnitListener#onSUDeployed(java.lang.String,
     *      java.lang.String,
     *      org.objectweb.petals.tools.jbicommon.descriptor.JBIDescriptor)
     */
    public void onSUDeployed(String serviceUnitName, String suRootPath,
        JBIDescriptor descriptor) throws PEtALSComponentSDKException {

        int providesCount = descriptor.getServices().getProvides().size();
        int consumesCount = descriptor.getServices().getConsumes().size();

        if (providesCount > 0) {
            // deploy the BPEL processes
            onSUDeployedBPELProcess(serviceUnitName, suRootPath, descriptor);
        }

        if (consumesCount > 0) {
            // activate new JMS connection
            onSUDeployedJMSProcess(serviceUnitName, suRootPath, descriptor);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.component.common.ServiceUnitListener#onSUStarted(java.lang.String)
     */
    public void onSUStarted(String serviceUnitName)
        throws PEtALSComponentSDKException {

        onSUStartedBPELProcess(serviceUnitName);
        onSUStartedJMSProcess(serviceUnitName);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.component.common.ServiceUnitListener#onSUStopped(java.lang.String)
     */
    public void onSUStopped(String serviceUnitName)
        throws PEtALSComponentSDKException {

        onSUStoppedBPELProcess(serviceUnitName);
        onSUStoppedJMSProcess(serviceUnitName);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.component.common.ServiceUnitListener#onSUUndeployed(java.lang.String,
     *      java.lang.String,
     *      org.objectweb.petals.tools.jbicommon.descriptor.JBIDescriptor)
     */
    public void onSUUndeployed(String serviceUnitName, String suRootPath,
        JBIDescriptor descriptor) throws PEtALSComponentSDKException {

        onSUUndeployedBPELProcess(serviceUnitName, suRootPath, descriptor);
        onSUUndeployedJMSProcess(serviceUnitName, suRootPath, descriptor);
    }

    /* ~~~ JMS processing ~~~ */

    /**
     * 
     * @param serviceUnitName
     * @param suRootPath
     * @param descriptor
     * @throws PEtALSComponentSDKException
     */
    private void onSUDeployedJMSProcess(String serviceUnitName,
        String suRootPath, JBIDescriptor descriptor)
        throws PEtALSComponentSDKException {

        Extensions extensions = descriptor.getServices().getConsumes().get(0)
            .getExtensions();

        // Create the JNDI context used to retrieve jms connection and
        // destination
        Context context = setupContext(suRootPath, extensions);

        // create the jmsConnection object that contains all information
        JMSConnection connection = setupConnection(context, extensions);

        // retrieve the jms 'address' to be used as key
        String address = PetalsExtensionsUtil
            .extractValueFromKeyValueExtension(extensions, ADDRESS);

        // store the jmsconnection
        jmsAddresses.put(serviceUnitName, address);
        jmsConsumerConnections.put(address, connection);
    }

    /**
     * 
     * @param address
     * @throws PEtALSComponentSDKException
     */
    private void onSUStartedJMSProcess(String serviceUnitName)
        throws PEtALSComponentSDKException {

        String address = jmsAddresses.get(serviceUnitName);

        if (address != null) {
            JMSConnection jmsConnection = jmsConsumerConnections.get(address);
            jmsConnection.start();
        }
    }

    /**
     * 
     * @param serviceUnitName
     * @param suRootPath
     * @param descriptor
     * @throws PEtALSComponentSDKException
     */
    private void onSUUndeployedJMSProcess(String serviceUnitName,
        String suRootPath, JBIDescriptor descriptor)
        throws PEtALSComponentSDKException {

        String address = jmsAddresses.remove(serviceUnitName);
        if (address != null) {
            jmsConsumerConnections.remove(address);
        }
    }

    /**
     * 
     * @param address
     * @throws PEtALSComponentSDKException
     */
    private void onSUStoppedJMSProcess(String serviceUnitName)
        throws PEtALSComponentSDKException {

        String address = jmsAddresses.get(serviceUnitName);

        if (address != null) {
            JMSConnection jmsConnection = jmsConsumerConnections.get(address);
            jmsConnection.stop();
        }

    }

    /* ~~~ BPEL processing ~~~ */

    /**
     * 
     * @param serviceUnitName
     * @param suRootPath
     * @param descriptor
     * @throws PEtALSComponentSDKException
     */
    private void onSUDeployedBPELProcess(String serviceUnitName,
        String suRootPath, JBIDescriptor descriptor)
        throws PEtALSComponentSDKException {

        Vector<File> externalWS = null;
        File wsdlFile = null;
        File bpelFile = null;

        // keep only bpel and wsdl files
        File[] processFiles = new File(suRootPath).listFiles(this.filter);
        if (processFiles == null || processFiles.length < 2) {
            throw new PEtALSComponentSDKException(
                "There must have at least one bpel file and one wsdl file");
        }

        // get process name
        String processName = findProcessName(processFiles);

        // get the BPEL file
        bpelFile = getBPELFile(processFiles, processName);

        // get external WSDL used by BPEL process
        externalWS = getExternalWSDL(processFiles, processName);

        wsdlFile = getWSDLFile(processFiles, processName);

        try {
            createBPELProcessInBsoap(processName, bpelFile, wsdlFile,
                externalWS);
        } catch (Throwable t) {
            throw new PEtALSComponentSDKException(t.getMessage(), t);
        }
    }

    private void onSUStartedBPELProcess(String address)
        throws PEtALSComponentSDKException {

    }

    /**
     * 
     * @param serviceUnitName
     * @param suRootPath
     * @param descriptor
     * @throws PEtALSComponentSDKException
     */
    private void onSUUndeployedBPELProcess(String serviceUnitName,
        String suRootPath, JBIDescriptor descriptor)
        throws PEtALSComponentSDKException {
        // remove BPEL processes from BSOAP server
        try {
            OrchestraConnection.connect();
            BpelAdminSession adminSession = OrchestraConnection
                .getBpelAdminSession();

            for (String processName : deployedProcesses) {
                logger.log(Level.FINE, "Undeploying BPEL process : "
                    + processName);
                adminSession.undeployBPELProcess(processName);
                adminSession.deleteBPELProcess(processName);
                deployedProcesses.remove(processName);
            }

        } catch (BPELAdminException e) {
            logger.log(Level.INFO, e.getMessage());
        } catch (OrchestraJbiException e) {
            logger.log(Level.INFO, e.getMessage());
        } catch (RemoteException e) {
            logger.log(Level.INFO, e.getMessage());
        }
    }

    private void onSUStoppedBPELProcess(String serviceUnitName)
        throws PEtALSComponentSDKException {

    }

    /* ~~~ JMS ~~~ */
    /**
     * Setup a JNDIContext to access a JMS destination.
     * 
     * @param suRootPath
     *            path where jndi context factory class can be found
     * @param extensions
     *            jndi properties
     * @return a JNDI context initialized with the given jbi.xml properties
     * @throws PEtALSComponentSDKException
     */
    private Context setupContext(String suRootPath, Extensions extensions)
        throws PEtALSComponentSDKException {

        // Create a class loader that will contains all JAR
        // It will load the specific JndiFactory class
        // to access the jndi containing Queue and Topic references
        URLClassLoader classLoader = createClassLoader(suRootPath, this
            .getClass().getClassLoader());

        // Create JNDI properties
        Properties jndiProperties = createJNDIProperties(extensions);

        // Create the JNDI context
        return createJNDIContext(classLoader, jndiProperties);
    }

    /**
     * extract jndi properties from the jbi.xml extensions
     * 
     * @param extensions
     * @return
     * @throws PEtALSComponentSDKException
     *             provider url or factory name not found
     */
    private Properties createJNDIProperties(Extensions extensions)
        throws PEtALSComponentSDKException {
        Properties outProperties = new Properties();

        String url = null;
        try {
            URI address = new URI(PetalsExtensionsUtil
                .extractValueFromKeyValueExtension(extensions, ADDRESS));
            url = address.getScheme() + ":" + address.getSchemeSpecificPart();
        } catch (URISyntaxException e) {
            throw new PEtALSComponentSDKException(ADDRESS
                + " property not found in the Service Unit descriptor file."
                + e);
        }
        String factory = PetalsExtensionsUtil
            .extractValueFromKeyValueExtension(extensions,
                INITIAL_CONTEXT_FACTORY);

        outProperties.setProperty(InitialContext.PROVIDER_URL, url);
        outProperties.setProperty(InitialContext.INITIAL_CONTEXT_FACTORY,
            factory);
        return outProperties;
    }

    /**
     * Return a JNDI context loaded from the JAR files contained in the
     * ServiceUnit and the Properties defined in the JBI.XML file.
     * 
     * @param classLoader
     *            contains the libraries where the JNDI context factory can be
     *            found
     * @param properties
     *            contains the factory class name and the url provider
     * @return a JNDI context
     * @throws PEtALSComponentSDKException
     *             can not retrieved the factory. properties might not be
     *             correct.
     */
    private Context createJNDIContext(ClassLoader classLoader,
        Properties properties) throws PEtALSComponentSDKException {
        Context ctx = null;
        try {
            InitialContextFactory icf = (InitialContextFactory) classLoader
                .loadClass(
                    properties
                        .getProperty(InitialContext.INITIAL_CONTEXT_FACTORY))
                .newInstance();
            ctx = icf.getInitialContext(properties);
        } catch (Exception e) {
            throw new PEtALSComponentSDKException(
                "JNDI context can not be created." + e);
        }
        return ctx;
    }

    /**
     * Setup a JMS connection to access a JMS destination.
     * 
     * @param context
     *            jndi context where the connection factory and the destination
     *            can be found
     * @param extensions
     *            contains connectionFactory and destination names, user/pwd,
     *            transacted and acknowledge modes.
     * @param role
     *            jms role : consumer or producer
     * @return a JMSConnection in a stopped state.
     * @throws PEtALSComponentSDKException
     *             a mandatory property is not found in the JBI.XML descriptor
     */
    private JMSConnection setupConnection(Context context, Extensions extensions)
        throws PEtALSComponentSDKException {
        String connName = PetalsExtensionsUtil
            .extractValueFromKeyValueExtension(extensions, CONNECTION_FACTORY);

        String destName = null;
        try {
            URI address = new URI(PetalsExtensionsUtil
                .extractValueFromKeyValueExtension(extensions, ADDRESS));
            destName = address.getFragment();
        } catch (URISyntaxException e) {
            throw new PEtALSComponentSDKException(ADDRESS
                + " property not found in the Service Unit descriptor file."
                + e);
        }

        String user = PetalsExtensionsUtil.extractValueFromKeyValueExtension(
            extensions, USER);

        String password = PetalsExtensionsUtil
            .extractValueFromKeyValueExtension(extensions, PASSWORD);

        boolean transacted = false;
        String transac = PetalsExtensionsUtil
            .extractValueFromKeyValueExtension(extensions, TRANSACTED);
        transacted = Boolean.valueOf(transac);

        return new JMSConnection(context, connName, destName, transacted, user,
            password);

    }

    /* ~~~ UTILITIES ~~~ */

    /**
     * Get the WSDL file of the BPEL process
     * 
     * @param processFiles
     * @param processName
     * @return
     */
    private File getWSDLFile(File[] processFiles, String processName) {
        File result = null;

        for (int i = 0; i < processFiles.length; i++) {
            if (processFiles[i].getName().equals(processName + ".wsdl")) {
                result = processFiles[i];
            }
        }
        return result;
    }

    /**
     * Get the external WSDL files (ie all the files where name is not
     * processName.wsdl)
     * 
     * @param processFiles
     * @param processName
     * @return
     */
    protected Vector<File> getExternalWSDL(File[] processFiles,
        String processName) {
        Vector<File> result = new Vector<File>();

        for (int i = 0; i < processFiles.length; i++) {
            File file = processFiles[i];
            String fileName = file.getName();
            if (fileName.endsWith(".wsdl")
                && !processName.equals(fileName.substring(0, fileName
                    .lastIndexOf('.')))) {
                result.add(file);
            }
        }
        return result;
    }

    /**
     * Get the BPEL file from all the process files
     * 
     * @param processFiles
     * @param processName
     * @return
     */
    private File getBPELFile(File[] processFiles, String processName)
        throws PEtALSComponentSDKException {
        File result = null;

        for (int i = 0; i < processFiles.length; i++) {
            if (processFiles[i].getName().endsWith(".bpel"))
                result = processFiles[i];
        }

        if (result == null)
            throw new PEtALSComponentSDKException("BPEL file (" + processName
                + ".bpel) has not been found");

        return result;
    }

    /**
     * Install in BSOAP
     * 
     * @param processName
     * @param bpelFile
     * @param wsdlFile
     * @param externalWS
     * @throws OrchestraJbiException
     * @throws JBIException
     * @throws RemoteException
     * @throws BPELAdminException
     */
    protected void createBPELProcessInBsoap(String processName, File bpelFile,
        File wsdlFile, Vector<File> externalWS) throws OrchestraJbiException,
        RemoteException, BPELAdminException {

        logger.log(Level.INFO, "Deploying BPEL process : " + processName);

        String[] exts = new String[externalWS.size()];
        for (int i = 0; i < externalWS.size(); i++) {
            exts[i] = externalWS.get(i).getAbsolutePath();
        }

        // connect to Bsoap
        OrchestraConnection.connect();
        BpelAdminSession adminSession = OrchestraConnection
            .getBpelAdminSession();

        // delete the process before deploying it
        adminSession.deleteBPELProcess(processName);

        // create and deploy the process
        adminSession.createProcess(processName, "bpeladmin");
        adminSession.deployProcess(processName, bpelFile.getAbsolutePath(),
            wsdlFile.getAbsolutePath(), exts, "", false);

        deployedProcesses.add(processName);
    }

    /**
     * Find the process name from the files. The process name is the name of the
     * bpel file without the bpel extension.
     * 
     * @param processFiles
     * @return
     * @throws DeploymentException
     */
    protected String findProcessName(File[] processFiles)
        throws PEtALSComponentSDKException {
        String processName = null;
        Vector<File> bpelFiles = new Vector<File>();

        // get all the bpel files
        for (int i = 0; i < processFiles.length; i++) {
            File tmp = processFiles[i];
            if (tmp.getName().endsWith(".bpel")) {
                bpelFiles.add(tmp);
            }
        }

        // more than one bpel file is not permitted !
        if (bpelFiles.size() == 0 || bpelFiles.size() > 1) {
            throw new PEtALSComponentSDKException(
                "There must have exactly one BPEL file provided", null);
        }

        // Get the process name from the file name
        File bpelFile = bpelFiles.get(0);
        String bpelFileName = bpelFile.getName();
        processName = bpelFileName.substring(0, bpelFileName.length()
            - ".bpel".length());

        return processName;
    }

    /**
     * A File name filter used to get only wsdl and bpel files.
     * 
     */
    public static class BpelFilter implements FilenameFilter {

        /*
         * (non-Javadoc)
         * 
         * @see java.io.FilenameFilter#accept(java.io.File, java.lang.String)
         */
        public boolean accept(File dir, String name) {
            boolean ret = true;
            ret = ret && dir.exists() && dir.isDirectory()
                && name.endsWith(".wsdl") || name.endsWith(".bpel");
            ret = ret && new File(dir + File.separator + name).exists();
            return ret;
        }
    }
}
