/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.engine.orchestra.util;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.rmi.RemoteException;

import javax.ejb.CreateException;
import javax.jbi.JBIException;
import javax.naming.NamingException;
import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;

import org.objectweb.orchestra.admin.BpelAdminSession;
import org.objectweb.orchestra.admin.BpelAdminSessionHome;
import org.objectweb.orchestra.admin.BpelAdminSessionUtil;
import org.objectweb.orchestra.util.BPELConstants;
import org.objectweb.orchestra.util.SimpleCallbackHandler;
import org.objectweb.petals.engine.orchestra.exception.OrchestraJbiException;

/**
 * Utilities to connect to the BSOAP server.
 * 
 * @author Christophe HAMERLING - eBMWebSourcing
 * 
 */
public class OrchestraConnection {

    /**
     * Connect to the orchestra server.
     * 
     * @throws OrchestraJbiException
     * @throws JBIException
     */
    public static void connect() throws OrchestraJbiException {

        SimpleCallbackHandler handler = new SimpleCallbackHandler(
            BPELConstants.USERNAME, BPELConstants.PASSWORD.toCharArray());

        // get current class loader
        ClassLoader threadCL = Thread.currentThread().getContextClassLoader();

        // set new class loader
        Thread.currentThread().setContextClassLoader(
            OrchestraConnection.class.getClassLoader());

        try {
            // configure required system properties
            File jonasBase = findJonasBaseDirectory();
            File jaasFile = new File(jonasBase, File.separator + "conf"
                + File.separator + "jaas.config");
            System.setProperty("java.security.auth.login.config", jaasFile
                .getAbsolutePath());
            System.setProperty("java.naming.factory.initial",
                "org.objectweb.carol.jndi.spi.MultiOrbInitialContextFactory");

            // login
            LoginContext lc = new LoginContext(BPELConstants.LOGIN_MODULE,
                handler);
            lc.login();

        } catch (LoginException e) {
            throw new OrchestraJbiException(
                "LoginException caught in OrchestraConnection.connectToOrchestra : ",
                e);
        } finally {
            // restore class loader
            Thread.currentThread().setContextClassLoader(threadCL);
        }
    }

    /**
     * Get the admin session
     * 
     * @return
     * @throws OrchestraJbiException
     * @throws JBIException
     */
    public static BpelAdminSession getBpelAdminSession()
        throws OrchestraJbiException {

        BpelAdminSession admin = null;
        // save current class loader
        ClassLoader old = Thread.currentThread().getContextClassLoader();

        // get the stub class loader
        Thread.currentThread().setContextClassLoader(
            getGeneratedStubClassLoader());

        try {
            // connect to administration module
            BpelAdminSessionHome home = BpelAdminSessionUtil.getHome();
            admin = home.create();

        } catch (NamingException e) {
            throw new OrchestraJbiException(
                "NamingException caught in OrchestraConnection.getOrchestraAdmin : ",
                e);
            
        } catch (CreateException e) {
            throw new OrchestraJbiException(
                "CreateException caught in OrchestraConnection.getOrchestraAdmin : ",
                e);
            
        } catch (RemoteException e) {
            throw new OrchestraJbiException(
                "RemoteException caught in OrchestraConnection.getOrchestraAdmin : ",
                e);
            
        } finally {
            // restore old class loader
            Thread.currentThread().setContextClassLoader(old);
        }
        return admin;
    }

    /**
     * Return a classloader for the orchestra.jar file which contains Orchestra
     * EJB's stubs. These jar is located under the JONAS_BASE directory.
     * 
     * @return the classloader that references the stubs.
     * @throws JBIException
     *             JONAS_BASE not set, or bpel.jar not found
     */
    public static ClassLoader getGeneratedStubClassLoader() throws OrchestraJbiException {

        // find JONAS_BASE directory
        File jonasbase = findJonasBaseDirectory();

        // find the stub jar in the BPEL work directory
        File tmp = new File(jonasbase + File.separator + "work"
            + File.separator + "apps" + File.separator + "BPEL");
        File stubJarFile = null;
        if (tmp.exists()) {
            for (File file : tmp.listFiles()) {
                if (file.getName().startsWith("orchestra")) {
                    stubJarFile = new File(file, "orchestra.jar");
                }
            }
        }
        
        if (stubJarFile == null) {
            throw new OrchestraJbiException(
                "Orchestra orchestra.jar stubs can not be found in the JONAS_BASE subdirectories.");
        }

        // create the class loader
        URL url;
        try {
            url = stubJarFile.toURL();
        } catch (MalformedURLException e) {
            throw new OrchestraJbiException(e);
        }
        URLClassLoader cl = new URLClassLoader(new URL[] {url},
            OrchestraConnection.class.getClassLoader());

        return cl;
    }

    /**
     * Return the File corresponding to the JONAS_BASE system variable. This
     * variable must be set !
     * 
     * @return
     * @throws JBIException
     *             JONAS_BASE var not found
     */
    protected static File findJonasBaseDirectory() throws OrchestraJbiException {
        String jonasBase = System.getProperty("jonas.base");
        if (jonasBase == null || jonasBase.trim().equals("")) {
            jonasBase = System.getenv("JONAS_BASE");
        }

        if (jonasBase == null) {
            throw new OrchestraJbiException(
                "JONAS_BASE is not defined! Set it in your system environment or start this JVM with the jonas.base property");
        }

        return new File(jonasBase);
    }
}
