/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: DeliveryChannelImpl.java,v 1.2 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.messaging;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import javax.jbi.JBIException;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.objectweb.petals.jbi.component.context.ComponentContextImpl;
import org.objectweb.petals.jbi.routing.Router;
import org.objectweb.petals.jbi.routing.RoutingException;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 * @author Gael BLONDELLE - EBM WebSourcing
 * @author ofabre - EBM Websourcing
 * @author rnaudin - EBM Websourcing
 */
public class DeliveryChannelImpl implements DeliveryChannel {

    public static final String PROPERTY_SENDSYNC = "javax.jbi.messaging.sendSync";

    public static final String PROPERTY_SENDSYNC_PROVIDER = "javax.jbi.messaging.sendSync.provider";

    public static final String PROPERTY_SENDSYNC_CONSUMER = "javax.jbi.messaging.sendSync.consumer";

    public static final String PROPERTY_NOACK = "org.objectweb.petals.messaging.noack";

    protected ComponentContextImpl context;

    protected LoggingUtil log;

    /**
     * Monolog Logger instance.
     */
    protected Logger logger;

    /**
     * This messageExchangeFactory is used to return
     * <code>MessageExchangeDecorator</code> on accept() methods.
     */
    protected MessageExchangeFactoryImpl messageExchangeFactory;

    protected boolean opened = true;

    protected Router router;

    /**
     * Map of MsgExchange that are blocked during a Synchronous send
     */
    protected Map<String, MessageExchangeDecorator> waitingExchanges;

    /**
     * Constructor
     * 
     * @param context
     * @param router
     * @param logger
     */
    public DeliveryChannelImpl(ComponentContextImpl context, Router router,
        Logger logger) {
        this.logger = logger;

        log = new LoggingUtil(logger);

        messageExchangeFactory = new MessageExchangeFactoryImpl(this, context
            .getAddress(), logger);

        this.router = router;
        this.context = context;

        waitingExchanges = new HashMap<String, MessageExchangeDecorator>();
    }

    /**
     * Return a MessageExchangeDecorator with 'consumer' or 'provider' observer
     * role (depending on the messageExchange getRole())
     * 
     * @see javax.jbi.messaging.DeliveryChannel#accept()
     */
    public MessageExchange accept() throws MessagingException {
        log.call();

        return poll(true, 0);
    }

    /**
     * @see javax.jbi.messaging.DeliveryChannel#accept(long)
     */
    public MessageExchange accept(long timeoutMS) throws MessagingException {
        log.call();

        return poll(false, timeoutMS);
    }

    /**
     * Poll for a recept message exchange.
     * 
     * @param block
     * @param timeoutMS
     * 
     * @return The received message exchange
     * 
     * @throws MessagingException
     */
    public MessageExchange poll(boolean block, long timeoutMS)
        throws MessagingException {
        log.start();

        checkDeliveryChannelIsOpened();

        MessageExchangeImpl msg = null;

        try {
            msg = router.receive(context, timeoutMS, block);
        } catch (RoutingException e) {
            e.printStackTrace();
        }

        MessageExchange result = null;
        if (msg != null) {
            result = messageExchangeFactory.createExchangeDecorator(msg, msg
                .getPattern());
        }

        log.end();

        return result;
    }

    /**
     * Check if the <code>DeliveryChannel</code> is opened
     * 
     * @throws MessagingException
     */
    public synchronized void checkDeliveryChannelIsOpened()
        throws MessagingException {
        if (!opened) {
            throw new MessagingException("DeliveryChannel is closed.");
        }
    }

    /**
     * Check if the <code>MessageExchange</code> is Ok (not null and not
     * terminated)
     * 
     * @throws MessagingException
     */
    public void checkMessageExchange(MessageExchange exchange)
        throws MessagingException {
        if (exchange == null) {
            throw new MessagingException("MessageExchange is null.");
        }
        MessageExchangeDecorator exchangeDecorator = (MessageExchangeDecorator) exchange;

        if (exchangeDecorator.getMessageExchange().isTerminated()) {
            throw new MessagingException("MessageExchange is terminated.");
        }
    }

    /**
     * This method wake up all threads that are waiting for a message. This
     * method deregister all endpoints for this component. FIXME this method
     * does not wakeup waiting threads anymore
     * 
     * @see javax.jbi.messaging.DeliveryChannel#close() TODO close:manage the
     *      received messages which are not consummed and are still in the
     *      incoming message queue.
     */
    public synchronized void close() throws MessagingException {
        log.call();

        checkDeliveryChannelIsOpened();

        /*
         * Notify all the pending 'SYNC_SEND' threads that the DeliveryChannel
         * is closing All the <code>MessageExchanges</code> contained in the
         * <code>waitingExchanges</code> are notified.
         */
        synchronized (waitingExchanges) {
            Collection<MessageExchangeDecorator> values = waitingExchanges
                .values();
            for (MessageExchangeDecorator messageExchangeDecorator : values) {
                messageExchangeDecorator.notifyAll();
            }
        }

        try {
            context.deregisterAllEndpoints();
        } catch (JBIException e) {
            throw new MessagingException(e);
        }

        opened = false;
    }

    /**
     * @see javax.jbi.messaging.DeliveryChannel#createExchangeFactory()
     */
    public MessageExchangeFactory createExchangeFactory() {
        log.call();

        MessageExchangeFactoryImpl mef = new MessageExchangeFactoryImpl(this,
            context.getAddress(), logger);

        return mef;
    }

    /**
     * @see javax.jbi.messaging.DeliveryChannel#createExchangeFactory(javax.xml.namespace.QName)
     */
    public MessageExchangeFactory createExchangeFactory(QName interfaceName) {
        log.call(interfaceName);

        MessageExchangeFactoryImpl mef = new MessageExchangeFactoryImpl(this,
            context.getAddress(), logger);

        mef.setDefaultInterfaceName(interfaceName);

        return mef;
    }

    /**
     * @see javax.jbi.messaging.DeliveryChannel#createExchangeFactory(javax.jbi.servicedesc.ServiceEndpoint)
     */
    public MessageExchangeFactory createExchangeFactory(ServiceEndpoint endpoint) {
        log.call(endpoint);

        MessageExchangeFactoryImpl mef = new MessageExchangeFactoryImpl(this,
            context.getAddress(), logger);

        mef.setDefaultServiceEndpoint(endpoint);

        return mef;
    }

    /**
     * @see javax.jbi.messaging.DeliveryChannel#createExchangeFactoryForService(javax.xml.namespace.QName)
     */
    public MessageExchangeFactory createExchangeFactoryForService(
        QName serviceName) {
        log.call(serviceName);

        MessageExchangeFactoryImpl mef = new MessageExchangeFactoryImpl(this,
            context.getAddress(), logger);

        mef.setDefaultServiceName(serviceName);

        return mef;
    }

    public synchronized boolean isOpened() {
        return opened;
    }

    // ///////////
    // non jbi
    // //////////

    /**
     * Used by the <code>Router</code> when a response message is received by
     * the <code>Transporter</code> and the recepient of the message is the
     * component attached to this channel. The message is a response of a
     * sendSync() method, the matching sendSync() must be unblocked with the
     * attached response.
     * 
     * @param msg
     *            the message received by the <code>Transporter</code>
     * @throws JBIException
     *             The channel has been closed by the component. The message can
     *             not be accepted
     */
    public boolean push(MessageExchangeImpl msg) throws JBIException {
        log.start(msg);

        boolean isMessageMatched = false;

        checkDeliveryChannelIsOpened();

        // Check if this ME is synchronized to a blocked sendSync().
        // If so, we unblock the sendSync() method, otherwise,
        // we return the message to the local topic, to be consumed by the next
        // accept().

        MessageExchangeDecorator decorator = waitingExchanges
            .get(waitingExchangeKey(msg));

        if (decorator != null) {
            isMessageMatched = true;
            // we notify the ProviderMEDecorator that its response is received.
            synchronized (decorator) {
                decorator.setMessageExchange(msg);
                decorator.setWaitingOnSynchronousSend(false);
                decorator.notifyAll();
            }
        }

        log.end();
        return isMessageMatched;
    }

    /**
     * @see javax.jbi.messaging.DeliveryChannel#send(javax.jbi.messaging.MessageExchange)
     */
    public void send(MessageExchange exchange) throws MessagingException {
        log.call(exchange);

        sendExchange(exchange, 0);
    }

    /**
     * @see javax.jbi.messaging.DeliveryChannel#sendSynch(javax.jbi.messaging.MessageExchange)
     */
    public boolean sendSync(MessageExchange exchange) throws MessagingException {
        log.call();
        return sendSync(exchange, 0);
    }

    /**
     * @see javax.jbi.messaging.DeliveryChannel#sendSynch(javax.jbi.messaging.MessageExchange,
     *      long)
     */
    public boolean sendSync(MessageExchange exchange, long timeoutMS)
        throws MessagingException {
        log.start();

        MessageExchangeDecorator exchangeDecorator = (MessageExchangeDecorator) exchange;

        synchronized (exchangeDecorator) {
            try {
                // Prepare the send
                exchangeDecorator.setWaitingOnSynchronousSend(true);

                // set the Properties to handle properly the SEND_SYNC mechanism
                setSendSyncProperties(exchangeDecorator);

                // Reference this exchange in the Map of waitingExchanges
                // It will be used to wake up the exchange when the response
                // will arrive
                waitingExchanges.put(waitingExchangeKey(exchangeDecorator),
                    exchangeDecorator);

                sendExchange(exchangeDecorator, timeoutMS);

                // wait for response or timeout
                exchangeDecorator.wait(timeoutMS);

                // Now, the response has been sent (wait unblocked),
                // or timeout occured.

                // First, unreference the exchange
                waitingExchanges.remove(waitingExchangeKey(exchangeDecorator));

                // Next, check if timeout or response
                if (exchangeDecorator.isWaitingOnSynchronousSend()) {
                    // we were always waiting for the response

                    // get the ownership of the ME
                    exchangeDecorator.getMessageExchange().setRole(
                        exchangeDecorator.getObserverRole());

                    exchangeDecorator.setStatus(ExchangeStatus.ERROR);

                    exchangeDecorator.getMessageExchange().setTerminated(true);
                }

            } catch (InterruptedException e) {
                log.error(e.getMessage(), e);
                // get the ownership of the ME
                exchangeDecorator.getMessageExchange().setRole(
                    exchangeDecorator.getObserverRole());

                exchangeDecorator.setStatus(ExchangeStatus.ERROR);

                exchangeDecorator.getMessageExchange().setTerminated(true);

                throw new MessagingException(e);
            }
        }
        log.end();

        return !exchangeDecorator.isWaitingOnSynchronousSend();
    }

    /**
     * Send the exchange to the underlying layer.<br>
     * Unused data in the exchange are cleaned before.<br>
     * If the message to exchange is a mere status information, and the exchange
     * is set to not send acknowledgment, the message is not sent.<br>
     * 
     * @see javax.jbi.messaging.DeliveryChannel#send(javax.jbi.messaging.MessageExchange)
     * @see org.objectweb.petals.jbi.messaging.MessageExchangeImpl#cleanMessages()
     */
    protected void sendExchange(MessageExchange exchange, long timeOut)
        throws MessagingException {
        log.start(exchange);

        checkDeliveryChannelIsOpened();
        checkMessageExchange(exchange);

        MessageExchangeDecorator exchangeDecorator = (MessageExchangeDecorator) exchange;
        MessageExchangeImpl exchangeImpl = exchangeDecorator
            .getMessageExchange();

        try {
            // clean unused messages
            exchangeImpl.cleanMessages();

            // check if the exchange must become terminated
            if (!ExchangeStatus.ACTIVE.equals(exchangeImpl.getStatus())) {
                exchangeImpl.setTerminated(true);
            }

            // check if the status information message must be send back
            boolean sendMessage = true;
            if (exchangeImpl.isTerminated()) {
                Object noAck = exchange.getProperty(PROPERTY_NOACK);
                sendMessage = (noAck == null || !noAck.toString().toLowerCase()
                    .equals("true"));
            }

            // send the message
            if (sendMessage) {
                router.send(context, exchangeImpl, timeOut);
            }

        } catch (JBIException e) {
            throw new MessagingException(e);
        }
        log.end();
    }

    /**
     * Set the properties to the message exchange to specify that a synchronious
     * response is expected
     * 
     * @param exchangeDecorator
     * 
     * @throws MessagingException
     *             The <code>sync_send()</code> is forbidden for
     *             acknowledgment send
     */
    protected void setSendSyncProperties(
        MessageExchangeDecorator exchangeDecorator) throws MessagingException {
        if (!exchangeDecorator.getStatus().equals(ExchangeStatus.ACTIVE)) {
            throw new MessagingException(
                "Synchronious send is forbidden when sending an acknwoledgements");
        }

        // specify the role of the requestor of the sync exchange, used
        // internally
        if (MessageExchange.Role.CONSUMER.equals(exchangeDecorator.getRole()))
            exchangeDecorator.setProperty(PROPERTY_SENDSYNC_CONSUMER, "true");
        else
            exchangeDecorator.setProperty(PROPERTY_SENDSYNC_PROVIDER, "true");

        // set the property as defined in the JBI specifications
        exchangeDecorator.setProperty(PROPERTY_SENDSYNC, "true");
    }

    /**
     * Return the string key associated with this exchange : role+exchangeId. It
     * is used to store synchronized exchanges. <br>
     * Note: if the exchange object is <code>MessageExchangeDecorator</code>,
     * we retreive the underlying <code>MessageExchangeImpl</code> object, to
     * avoid the checks made on getExchangeId() or getRole().
     * 
     * @param exchange
     * @return
     */
    protected String waitingExchangeKey(MessageExchange exchange) {
        MessageExchangeImpl meImpl = null;

        if (exchange instanceof MessageExchangeDecorator) {
            meImpl = ((MessageExchangeDecorator) exchange).getMessageExchange();
        } else {
            meImpl = (MessageExchangeImpl) exchange;
        }
        return meImpl.getRole() + meImpl.getExchangeId();
    }
}
