/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: MessageExchangeSerializer.java 154 2006-03-27 15:30:10Z alouis $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.messaging;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.ArrayBlockingQueue;

import javax.activation.DataHandler;
import javax.mail.util.ByteArrayDataSource;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.objectweb.petals.util.LoggingUtil;

/**
 * 
 * @author alouis - EBM Websourcing
 * 
 */
public class MessageExchangeSerializer implements Serializable {

    private static final long serialVersionUID = 1L;

    private static int POOLSIZE = 8;

    private LoggingUtil logger;

    private ArrayBlockingQueue<Transformer> transformerPool;

    private MessageExchangeSerializer() {
        super();
        // create the transformer pool
        transformerPool = new ArrayBlockingQueue<Transformer>(POOLSIZE);
        for (int i = 0; i < POOLSIZE; i++) {
            transformerPool.add(createNewTransformer());
        }
    }

    protected static MessageExchangeSerializer instance = new MessageExchangeSerializer();

    public static MessageExchangeSerializer instance() {
        return instance;
    }

    public void serializeContent(Source content, ObjectOutputStream s,
        boolean compress) throws Exception {
        synchronized (content) {
            /*
             * Compress content if specified
             */
            StreamResult sresult = null;
//            if (compress) {
//                GZIPOutputStream gz = new GZIPOutputStream(s);
//                sresult = new StreamResult(gz);
//            } else {
                sresult = new StreamResult(s);
//            }
            Transformer transform = transformerPool.take();
            transform.transform(content, sresult);
            transformerPool.offer(transform);

        }
    }

    public Source deserializeContent(ObjectInputStream s, boolean compress)
        throws IOException {
        StreamSource source = new StreamSource();

        InputStream in = s;
//        if (compress) {
//            in = new GZIPInputStream(s);
//        }

        byte[] buff = new byte[1024];
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        int n = 0;
        while (-1 != (n = in.read(buff))) {
            outputStream.write(buff, 0, n);
        }

        ByteArrayInputStream result = new ByteArrayInputStream(outputStream
            .toByteArray());

        source.setInputStream(result);
        return source;
    }

    /**
     * Serialize all datahandler in the Map. <br>
     * Serialized information is : <br>
     * -attachmentName <br>
     * -attachmentContentType <br>
     * -name <br>
     * -inputStream available bytes count <br>
     * -inputStream as a byteArray
     * 
     * @param attachments
     * @param s
     * @throws Exception
     */
    public void serializeAttachments(Map<String, DataHandler> attachments,
        ObjectOutputStream s) throws Exception {
        s.writeInt(attachments.size());

        for (Iterator iter = attachments.keySet().iterator(); iter.hasNext();) {

            String attachmentName = (String) iter.next();
            DataHandler attachment = attachments.get(attachmentName);

            InputStream binary = attachment.getInputStream();
            synchronized (binary) {

                int available = binary.available();

                // write information about the attachment
                s.writeUTF(attachmentName);
                s.writeUTF(attachment.getContentType());
                s.writeUTF(attachment.getName());
                s.writeInt(available);

                // write attachment content
                byte[] content = new byte[available];
                binary.read(content);
                s.write(content);
                s.flush();
            }
        }
    }

    /**
     * Deserialize all datahandlers in the stream and return a Map. <br>
     * Deserialized information is : <br>
     * -attachmentName <br>
     * -attachmentContentType <br>
     * -name <br>
     * -Stream available bytes count <br>
     * -Stream as a byteArray
     * 
     * @param s
     * @return
     * @throws IOException
     */
    public Map<String, DataHandler> deserializeAttachments(ObjectInputStream s)
        throws IOException {
        HashMap<String, DataHandler> attachments = new HashMap<String, DataHandler>();

        int attachmentsCount = s.readInt();

        for (int i = 0; i < attachmentsCount; i++) {
            // read information
            String attachmentName = s.readUTF();
            String contentType = s.readUTF();
            String name = s.readUTF();
            int available = s.readInt();

            // read content
            byte[] content = new byte[available];
            s.readFully(content);
            ByteArrayDataSource dataSource = new ByteArrayDataSource(content,
                contentType);
            dataSource.setName(name);

            // create the dataHandler
            DataHandler dataHandler = new DataHandler(dataSource);

            attachments.put(attachmentName, dataHandler);
        }
        return attachments;
    }

    private Transformer createNewTransformer() {
        Transformer transformer = null;
        try {
            transformer = TransformerFactory.newInstance().newTransformer();
        } catch (TransformerConfigurationException e) {
            logger.error("Can't create transformer pool", e);
        } catch (TransformerFactoryConfigurationError e) {
            logger.error("Can't create transformer pool", e);
        }
        Properties props = new Properties();
        props.put(OutputKeys.OMIT_XML_DECLARATION, "yes");
        if (transformer != null) {
            transformer.setOutputProperties(props);
        }
        return transformer;
    }

    public void setLogger(LoggingUtil logger) {
        this.logger = logger;
    }

}
