/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: RouterImpl.java,v 1.2 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.routing;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessageExchange;

import org.objectweb.fractal.fraclet.annotation.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.Interface;
import org.objectweb.fractal.fraclet.annotation.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.LifeCycleType;
import org.objectweb.fractal.fraclet.annotation.Monolog;
import org.objectweb.fractal.fraclet.annotation.Provides;
import org.objectweb.fractal.fraclet.annotation.Requires;
import org.objectweb.petals.jbi.component.context.ComponentContextImpl;
import org.objectweb.petals.jbi.component.lifecycle.ComponentLifeCycle;
import org.objectweb.petals.jbi.management.endpoint.EndpointService;
import org.objectweb.petals.jbi.management.lifecycle.LifeCycleManagerService;
import org.objectweb.petals.jbi.messaging.DeliveryChannelImpl;
import org.objectweb.petals.jbi.messaging.MessageExchangeImpl;
import org.objectweb.petals.jbi.registry.AbstractEndpoint;
import org.objectweb.petals.jbi.registry.ConsumerEndpoint;
import org.objectweb.petals.jbi.registry.InternalEndpoint;
import org.objectweb.petals.jbi.transport.TransportException;
import org.objectweb.petals.jbi.transport.Transporter;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * Routes messages to their destinations. Performs an itinerary resolution,
 * adding for example a transformatino service before delivering the message to
 * its recipient, and perform an address resolution, finding the endpoint
 * satisfying the the destination information (finding an endpoint for a
 * specified service). Resolutions are based on defined rules.
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.objectweb.petals.jbi.routing.Router.class))
public class RouterImpl implements Router {

    protected static final long TIMEOUT = 10;

    @Requires(name = "endpoint-service", signature = org.objectweb.petals.jbi.management.endpoint.EndpointService.class)
    protected EndpointService endpointService;

    /**
     * logger wrapper
     */
    protected LoggingUtil log;

    @Monolog(name = "logger")
    protected Logger logger;

    @Requires(name = "lifecyclemanager", signature = org.objectweb.petals.jbi.management.lifecycle.LifeCycleManagerService.class)
    protected LifeCycleManagerService manager;

    @Requires(name = "transporter", signature = org.objectweb.petals.jbi.transport.Transporter.class)
    protected Transporter transporter;

    protected AddressResolver addressResolver;

    // ----------------------------------------------------------------------
    // Fractal BindingController implementation
    // ----------------------------------------------------------------------

    /**
     * Sends the specified message. Depending on the Role of the sender, send
     * the message to the producer (request) or to the consumer (response)
     * 
     * @param source
     * @param exchange
     * @param timeOut
     * @throws RoutingException
     */
    public void send(ComponentContextImpl source, MessageExchangeImpl exchange,
        long timeOut) throws RoutingException {
        log.call();

        if (MessageExchange.Role.CONSUMER.equals(exchange.getRole())) {
            sendToProvider(source, exchange, timeOut);
        } else {
            sendToConsumer(source, exchange, timeOut);
        }
    }

    // ----------------------------------------------------------------------
    // Fractal LifecycleController implementation
    // ----------------------------------------------------------------------

    @LifeCycle(on = LifeCycleType.START)
    protected void start() {
        log = new LoggingUtil(logger);
        log.start();
        addressResolver = new AddressResolver(endpointService);
        log.end();
    }

    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() {
        log.start();

        logger = null;

        log.end();
    }

    /**
     * Sends the specified message, sent by the specified component. TODO send
     * directly to the DeliveryChannel of the component related to the recipient
     * endpoint.
     * 
     * @param source
     * @param exchange
     * @param timeOut
     * @throws RoutingException
     */
    protected void sendToConsumer(ComponentContextImpl source,
        MessageExchangeImpl exchange, long timeOut) throws RoutingException {
        log.start();

        try {
            AbstractEndpoint consumerEndpoint = exchange.getConsumerEndpoint();

            exchange.setRole(MessageExchange.Role.CONSUMER);

            transporter.send(exchange, consumerEndpoint.getContainerName(),
                consumerEndpoint.getComponentName(), timeOut);
        } catch (TransportException e) {
            // restore the Role if the message has not been sent
            exchange.setRole(MessageExchange.Role.PROVIDER);

            throw new RoutingException(e);
        }

        log.end();
    }

    /**
     * Sends the specified message to the provider of the service. TODO for the
     * moment, send directly to the DeliveryChannel of the component related to
     * the recipient endpoint.
     * 
     * @param source
     * @param exchange
     * @throws RoutingException
     */
    protected void sendToProvider(ComponentContextImpl source,
        MessageExchangeImpl exchange, long timeOut) throws RoutingException {
        log.start();

        // resolve the address (find endpoint)
        addressResolver.resolveAddress(source, exchange);

        try {
            AbstractEndpoint providerEndpoint = (AbstractEndpoint) exchange
                .getEndpoint();

            exchange.setRole(MessageExchange.Role.PROVIDER);

            transporter.send(exchange, providerEndpoint.getContainerName(),
                providerEndpoint.getComponentName(), timeOut);
        } catch (TransportException e) {
            // restore the Role if the message has not been sent
            exchange.setRole(MessageExchange.Role.CONSUMER);

            // TODO set message to Error ?
            throw new RoutingException(e);
        }

        log.end();
    }

    /**
     * Receive a message from the Transporter and return it to the Delivery
     * Channel.
     * 
     * @param source
     *            The source component context
     * @param timeoutMS
     *            The timeout to receive the incomming message
     * @param block
     *            If <code>True</code>, the pol is blocking until a response
     */
    public MessageExchangeImpl receive(ComponentContextImpl source,
        long timeoutMS, boolean block) throws RoutingException {
        log.start();

        MessageExchangeImpl exchange = null;
        String componentName = source.getComponentName();

        try {
            exchange = transporter.poll(componentName, timeoutMS, block);
        } catch (TransportException e) {
            throw new RoutingException(e);
        }

        log.end();

        return exchange;
    }

    /**
     * Push a message from the MNR to the <code>DeliveryChannel</code>.
     * 
     * @param exchange
     *            The message exchange to push
     * 
     * @throws RoutingException
     */
    public void pushResponse(MessageExchangeImpl exchange)
        throws RoutingException {
        log.start();

        // Find the name of the component to deliver the Exchange,
        // depending on the Role (consumer or provider)
        String componentName = null;

        if (MessageExchange.Role.PROVIDER.equals(exchange.getRole())) {
            componentName = ((InternalEndpoint) exchange.getEndpoint())
                .getComponentName();
            // remove the SENDSYNC property for eventually futher exchanges
            exchange.setProperty(DeliveryChannelImpl.PROPERTY_SENDSYNC, null);
            exchange.setProperty(
                DeliveryChannelImpl.PROPERTY_SENDSYNC_PROVIDER, null);
        } else {
            componentName = ((ConsumerEndpoint) exchange.getConsumerEndpoint())
                .getComponentName();
            // remove the SENDSYNC property for eventually futher exchanges
            exchange.setProperty(DeliveryChannelImpl.PROPERTY_SENDSYNC, null);
            exchange.setProperty(
                DeliveryChannelImpl.PROPERTY_SENDSYNC_CONSUMER, null);
        }

        // retrieve the ComponentLifeCycle
        ComponentLifeCycle comp = manager.getComponentByName(componentName);

        // Push the response message exchange
        try {
            if (!comp.getComponentContext().getDeliveryChannelImpl().push(
                exchange)) {
                // Resend the message to the local topic, it would be consumed
                // by a next accept()
                if (MessageExchange.Role.PROVIDER.equals(exchange.getRole()))
                    exchange.setRole(MessageExchange.Role.CONSUMER);
                else
                    exchange.setRole(MessageExchange.Role.PROVIDER);
                send(comp.getComponentContext(), exchange, TIMEOUT);
            }
        } catch (JBIException e) {
            throw new RoutingException(e);
        }
    }
}
