/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: JoramConnection.java,v 0.3 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.transport;

import java.net.ConnectException;
import java.net.UnknownHostException;
import java.util.Collection;
import java.util.Hashtable;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.Semaphore;

import javax.jms.Connection;
import javax.jms.DeliveryMode;
import javax.jms.ExceptionListener;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageListener;
import javax.jms.MessageProducer;
import javax.jms.Session;
import javax.jms.TopicSubscriber;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.objectweb.joram.client.jms.ConnectionFactory;
import org.objectweb.joram.client.jms.Topic;
import org.objectweb.joram.client.jms.admin.AdminException;
import org.objectweb.joram.client.jms.admin.AdminModule;
import org.objectweb.joram.client.jms.local.LocalConnectionFactory;
import org.objectweb.petals.jbi.messaging.MessageExchangeImpl;
import org.objectweb.petals.util.JNDIUtil;

/**
 * @author Adrien LOUIS - EBM WebSourcing
 * @author rnaudin - EBM WebSourcing
 * 
 */
public class JoramConnection {

    /**
     * Size of the send <code>Session</code> pool
     */
    private static int POOLSIZE = 8;

    /**
     * Prefix for the topic names
     */
    protected static final String TOPIC_NAME = "JoramConnection-";

    /**
     * Message Selector expression of the <code>SYNC_MODE</code> subscribers
     */
    protected static final String SYNCMODE_MSGSELECTOR = JoramTransporter.SYNC_MODE
        + " = TRUE ";

    /**
     * Message Selector expression of the 'normal' subscribers
     */
    public static final String NORMAL_MSGSELECTOR = JoramTransporter.SYNC_MODE
        + " IS NULL";

    protected String host;

    protected InitialContext ictx;

    protected String containerName;

    protected String pwd;

    protected Connection connection;

    /**
     * Pool of <code>Session</code> used to publish JMS messages to
     * destination topics
     */
    protected ArrayBlockingQueue<Session> sessionsPool;

    /**
     * Table of <code>Session</code> used to poll incoming messages.
     * componentName : <code>Session</code>
     */
    protected Hashtable<String, Session> pollSessionsTable = new Hashtable<String, Session>();

    /**
     * Table of <code>Session</code> used to listen incoming
     * <code>SEND_SYNC</code> responses messages. componentName :
     * <code>Session</code>
     */
    protected Hashtable<String, Session> listenerSessionsTable = new Hashtable<String, Session>();

    /**
     * Table of <code>TopicSubscriber</code> used to receive the 'normal'
     * messages Name of component : <code>TopicSubscriber</code>
     */
    protected Hashtable<String, TopicSubscriber> pollSubscribersTable = new Hashtable<String, TopicSubscriber>();

    /**
     * Table of <code>TopicSubscriber</code> used to receive the synchronious
     * responses messages. Name of component : <code>TopicSubscriber</code>
     */
    protected Hashtable<String, TopicSubscriber> listenerSubscribersTable = new Hashtable<String, TopicSubscriber>();

    protected Hashtable<String, Semaphore> semaphoresTable = new Hashtable<String, Semaphore>();

    /**
     * The <code>MessageListener</code> instance used by the listeners
     * subscribers in this connection
     */
    protected MessageListener messageListener;

    protected int tcp;

    protected String user;

    protected boolean connectionStarted = false;

    protected Serializer serializer = new ObjectSerializer();

    public JoramConnection(String containerName, int tcp, String user,
        String pwd, String host, InitialContext ictx) {
        this.containerName = containerName;
        this.tcp = tcp;
        this.user = user;
        this.pwd = pwd;
        this.host = host;
        this.ictx = ictx;
    }

    /**
     * Start the JMS connection and initialize the pool of send
     * <code>Session</code>
     * 
     * @param exceptionListener
     *            The exception listener of this connection. Used to receive
     *            asynchronious JMS errors.
     * @param messageListener
     *            the message listener of this connection. Used to receive
     *            asynchroniously the <code>SEND_SYNC</code> reponses
     *            messages.
     * @throws JMSException
     * @throws ConnectException
     */
    public void start(ExceptionListener exceptionListener,
        MessageListener messageListener) throws JMSException, ConnectException {

        ConnectionFactory connectionFactory = null;
        connectionFactory = (ConnectionFactory) LocalConnectionFactory.create();

        // FIXME optim multithread
        // connectionFactory.getParameters().multiThreadSync = true;
        // connectionFactory.getParameters().multiThreadSyncDelay = 1;

        connection = connectionFactory.createConnection(user, pwd);
        connection.setExceptionListener(exceptionListener);

        // create the sessions pool to send messages
        sessionsPool = new ArrayBlockingQueue<Session>(POOLSIZE);
        for (int i = 0; i < POOLSIZE; i++) {
            sessionsPool.add(connection.createSession(false,
                Session.CLIENT_ACKNOWLEDGE));
        }

        this.messageListener = messageListener;

        connection.start();
        connectionStarted = true;
    }

    /**
     * Close jms and jndi connections. Clean send <code>Session</code> pool.
     * Topics subscriptions are not removed.
     * 
     * @throws JMSException
     *             can not close the jms connection or session
     * @throws NamingException
     *             can not close the jndi context access
     */
    public void stop() throws JMSException, NamingException {
        if (connectionStarted) {
            connection.stop();
            connectionStarted = false;
        }

        connection.close();
        connectionStarted = false;

        Collection<Semaphore> values = semaphoresTable.values();
        // Unlock all the pending receive threads
        for (Semaphore semaphore : values) {
            semaphore.release();
        }
    }

    /**
     * Create a topic to receive the component messages and set 2 subscribers on
     * it; one for 'normal' messages and one for response of synchronious sends (<code>SEND_SYNC</code>).
     * 
     * @param componentName
     *            The name of the component
     * 
     * @throws AdminException
     * @throws NamingException
     * @throws ConnectException
     * @throws AdminException
     * @throws UnknownHostException
     */

    public void createTopicDevice(String componentName)
        throws ConnectException, NamingException, AdminException,
        UnknownHostException, JMSException {

        Topic localTopic = bindTopic(componentName);

        // retrieve the logical topic Name
        String topicName = getTopicName(containerName, componentName);

        // create the device for 'normal' messages
        Session session = connection.createSession(false,
            Session.AUTO_ACKNOWLEDGE);
        TopicSubscriber subscriber = session.createDurableSubscriber(
            localTopic, topicName, NORMAL_MSGSELECTOR, false);

        pollSessionsTable.put(topicName, session);
        pollSubscribersTable.put(topicName, subscriber);

        // Create the device for SEND_SYNC reponses messages
        session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);
        subscriber = session.createDurableSubscriber(localTopic, topicName
            + JoramTransporter.SYNC_MODE, SYNCMODE_MSGSELECTOR, false);
        subscriber.setMessageListener(messageListener);

        listenerSessionsTable.put(topicName, session);
        listenerSubscribersTable.put(topicName, subscriber);

        // Add the bound semaphore to the semaphore table
        semaphoresTable.put(topicName, new Semaphore(1));
    }

    /**
     * Create and bind the topic for this server in the global JNDI (create or
     * retrieve it). Connection to the Joram AdminModule is done internally.
     * 
     * @param componentName
     *            The name of the component
     * @return
     * @throws NamingException
     * @throws ConnectException
     * @throws AdminException
     * @throws UnknownHostException
     */
    protected Topic bindTopic(String componentName) throws NamingException,
        ConnectException, AdminException, UnknownHostException {
        Topic topic = null;
        String topicName = getTopicName(containerName, componentName);

        boolean alreadyBound = JNDIUtil.isBound(ictx, "/", topicName);

        if (alreadyBound) {
            // retrieve the existing topic
            topic = (Topic) ictx.lookup(topicName);
        } else {
            // create and bind a new topic
            connectAdminModule();
            topic = Topic.create(topicName);
            topic.setFreeReading();
            topic.setFreeWriting();
            disconnectAdminModule();
            ictx.bind(topicName, topic);
        }

        return topic;
    }

    /**
     * Delete and unbind the topic for this server in the global JNDI.
     * Connection to the Joram AdminModule is done internally.
     * 
     * @param componentName
     *            The name of the component
     * @return
     * @throws NamingException
     * @throws ConnectException
     * @throws UnknownHostException
     * @throws AdminException
     * @throws JMSException
     */
    protected void unbindTopic(String topicName) throws NamingException,
        ConnectException, UnknownHostException, AdminException, JMSException {
        boolean alreadyBound = JNDIUtil.isBound(ictx, "/", topicName);
        if (alreadyBound) {
            // Topic topic = (Topic) ictx.lookup(topicName);

            // Unbind the topic
            ictx.unbind(topicName);

            // Delete the existing topic
            /*
             * TODO rnaudin : make it works. connectAdminModule();
             * topic.delete(); disconnectAdminModule();
             */
        }
    }

    /**
     * Send the MessageExchange to the destination of the given component. The
     * exchange is serialized into a jms Message.
     * 
     * @param containerName
     *            The name of the destination container
     * @param componentName
     *            The name of the destination component
     * @param messageExchange
     *            the message exchange to serialized to a
     *            <code>JMS.ObjectMessage</code>
     * @param timeToLive
     * @param persistence
     *            Persistence or not of this transfer
     * @param isSyncMode
     *            SYNC_MODE is activated or not
     * @throws TransportException
     */
    public void sendToDestination(String containerName, String componentName,
        MessageExchangeImpl messageExchange, long timeToLive,
        boolean persistence, boolean isSyncMode) throws TransportException {

        Session distantSession = null;
        MessageProducer producer;

        try {
            // find the topic destination in jndi
            Topic remoteTopic = findTopic(containerName, componentName);

            // retrieve a session - block until a session is available
            distantSession = sessionsPool.take();

            // create a producer
            producer = distantSession.createProducer(remoteTopic);

            // serialise MessageExchange object to a JMS Object Message
            Message msg = serializer.jbi2jms(messageExchange, distantSession);

            // set persistence to true or false
            int persistenceMode = (persistence) ? DeliveryMode.PERSISTENT
                : DeliveryMode.NON_PERSISTENT;

            // set the SYNC_MODE JMS property to true
            if (isSyncMode) {
                msg.setBooleanProperty(JoramTransporter.SYNC_MODE, true);
            }

            producer.send(msg, persistenceMode, 0, timeToLive);

            producer.close();

            sessionsPool.offer(distantSession);

            distantSession = null;

        } catch (InterruptedException e) {
            throw new TransportException(
                "Problem accessing the JMS sessions pool.", e);
        } catch (JMSException e) {
            throw new TransportException(
                "Can not send the message to the JMS destination.", e);
        } finally {
            // release the session
            if (distantSession != null) {
                sessionsPool.offer(distantSession);
            }
        }
    }

    /**
     * Poll for a message on the Topic of the given component. The poll can be
     * blocking or instantaneous
     * 
     * @param componentName
     * @param timeoutMS
     * @param block
     * 
     * @return The retrieved message, can be <code>null</code>
     * 
     * @throws JMSException
     * @throws InterruptedException
     */
    public Message receiveForDestination(String componentName, long timeoutMS,
        boolean block) throws JMSException, InterruptedException {
        Message message;
        String topicName = getTopicName(containerName, componentName);
        TopicSubscriber subscriber = pollSubscribersTable.get(topicName);

        /*
         * Lock the component subscriber it has a critical section in the
         * receive() methods The synchronized() is not used as the critical
         * section of the subscriber must be reachable when closing the
         * connection.
         */
        Semaphore semaphore = semaphoresTable.get(topicName);
        semaphore.acquire();

        // Test if the thread has been awaiken by the closing of the connection
        if (connectionStarted == false) {
            // Each awaken thread would awake the next pending one
            if (semaphore.hasQueuedThreads())
                semaphore.release();
            return null;
        }

        if (block == true)
            message = subscriber.receive();
        else if (timeoutMS == 0)
            message = subscriber.receiveNoWait();
        else
            message = subscriber.receive(timeoutMS);

        semaphore.release();

        return message;
    }

    /**
     * Remove the subscribers and sessions bound to the component'Topic. The
     * topic is removed from the Joram configuration and from the Jndi server.
     * 
     * @param componentName
     * 
     * @throws JMSException
     * @throws NamingException
     * @throws ConnectException
     * @throws UnknownHostException
     * @throws AdminException
     */
    public void deleteTopicDevice(String componentName) throws JMSException,
        NamingException, ConnectException, UnknownHostException, AdminException {

        String topicName = getTopicName(containerName, componentName);

        pollSubscribersTable.remove(topicName);
        Session session = pollSessionsTable.remove(topicName);
        session.close();

        listenerSubscribersTable.remove(topicName);
        session = listenerSessionsTable.remove(topicName);
        session.setMessageListener(null);
        session.close();

        unbindTopic(topicName);
    }

    /**
     * Find in the JNDI context the registered Topic associated to the specified
     * containerName and componentName. The topic can be either local or remote.
     * 
     * @param componentName
     * @param containerName
     * 
     * @return the topic related to the containerName
     * 
     * @throws TransportException
     *             topic not found, or jndi access exception
     */
    protected Topic findTopic(String containerName, String componentName)
        throws TransportException {
        Topic remoteTopic = null;

        try {
            remoteTopic = (Topic) ictx.lookup(getTopicName(containerName,
                componentName));
        } catch (NamingException e) {
            throw new TransportException(
                "Can not find the Topic associate to the destination component '"
                    + componentName + "'", e);
        }

        return remoteTopic;
    }

    /**
     * Build the logical topic name according to the container and component
     * names.
     * 
     * @param containerName
     * @param componentName
     * 
     * @return the logical name of the topic
     */
    private static String getTopicName(String containerName,
        String componentName) {
        StringBuilder topicNameBuilder = new StringBuilder(TOPIC_NAME);
        topicNameBuilder.append(containerName).append("-")
            .append(componentName);

        return topicNameBuilder.toString();
    }

    /**
     * @see AdminModule#connect(String, int, String, String, int, String)
     */
    protected void connectAdminModule() throws ConnectException,
        UnknownHostException, AdminException {
        AdminModule.connect(host, tcp, user, pwd, 3);
    }

    /**
     * @see AdminModule#disconnect()
     */
    protected void disconnectAdminModule() {
        AdminModule.disconnect();
    }

}
