/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: JoramTransporter.java,v 0.3 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.transport;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessageExchange;
import javax.jms.ExceptionListener;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageListener;
import javax.jms.ObjectMessage;
import javax.naming.InitialContext;

import org.objectweb.fractal.fraclet.annotation.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.Interface;
import org.objectweb.fractal.fraclet.annotation.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.LifeCycleType;
import org.objectweb.fractal.fraclet.annotation.Monolog;
import org.objectweb.fractal.fraclet.annotation.Provides;
import org.objectweb.fractal.fraclet.annotation.Requires;
import org.objectweb.petals.jbi.messaging.DeliveryChannelImpl;
import org.objectweb.petals.jbi.messaging.MessageExchangeImpl;
import org.objectweb.petals.jbi.routing.Router;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.PropertyUtil;
import org.objectweb.petals.util.SystemUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * Implementation of the <code>Transporter</code> with JORAM.<br>
 * This implementation uses the fractal JoramAgent. <br>
 * For each component installed, the<code>JoramTransporter</code> creates and
 * registers a <code>Topic</code> in the distributed JNDI server with the
 * following name: <em>JoramConnection-[containerName]-[componentName]</em><br>
 * The name of a container is the <em>ID</em> specified in the
 * <em>conf/server.properties</em> file. <br>
 * Similary, at each uninstallation of component, the <code>Topic</code> is
 * deleted and unregistered. <br>
 * This <code>JoramTransporter</code> uses an <code>ObjectSerializer</code>
 * to create <em>jms.ObjectMessage</em> to send via JMS. So, the
 * <code>MessageExchange</code> to send is serialized.
 * 
 * @version $Rev: 477 $ $Date: 2006-05-29 17:18:07 +0200 (lun., 29 mai 2006) $
 * @since Petals 1.0
 * @author alouis
 * @author rnaudin
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.objectweb.petals.jbi.transport.Transporter.class))
public class JoramTransporter implements Transporter {

    public static final String SYNC_MODE = "syncMode";

    protected String host;

    protected String containerName;

    protected JoramConnection joramConnection;

    protected JoramAgent joramAgent;

    protected Serializer serializer = new ObjectSerializer();

    protected LoggingUtil log;

    protected String pwd;

    @Monolog(name = "logger")
    protected Logger logger;

    @Requires(name = "router", signature = org.objectweb.petals.jbi.routing.Router.class)
    protected Router router;

    protected int tcp;

    protected String user;

    /**
     * Send the message to the specified distant or local container, by
     * serializing the exchange with a <code>Serializer</code>. We retrieve
     * the related distant <code>Topic</code> with the name of the container,
     * transform the <code>MessageExchangeImpl</code> into a
     * <code>jms.Message</code>, and send this message onto the distant
     * topic.
     */
    public void send(MessageExchangeImpl messageExchange, String containerName,
        String componentName, long timeToLive) throws TransportException {
        log.start();
        boolean syncMode = false;

        // check if the message has to be persisted
        Object noPersist = messageExchange.getProperty(PROPERTY_NOPERSISTANCE);
        boolean persist = (noPersist == null || !noPersist.toString()
            .toLowerCase().equals("true"));

        if (MessageExchange.Role.CONSUMER.equals(messageExchange.getRole())) {
            syncMode = (messageExchange
                .getProperty(DeliveryChannelImpl.PROPERTY_SENDSYNC_CONSUMER) != null && ((String) messageExchange
                .getProperty(DeliveryChannelImpl.PROPERTY_SENDSYNC_CONSUMER))
                .toLowerCase().equals("true"));
        } else {
            syncMode = (messageExchange
                .getProperty(DeliveryChannelImpl.PROPERTY_SENDSYNC_PROVIDER) != null && ((String) messageExchange
                .getProperty(DeliveryChannelImpl.PROPERTY_SENDSYNC_PROVIDER))
                .toLowerCase().equals("true"));
        }

        try {
            joramConnection.sendToDestination(containerName, componentName,
                messageExchange, timeToLive, persist, syncMode);
        } catch (TransportException e) {
            log.error(e.getMessage(), e);
            throw e;
        }
        log.end();
    }

    /**
     * Initialise variables with joram.properties file, start the JORAM server
     * and and instanciate a <code>JoramConnection</code>
     * 
     * @throws Exception
     */
    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws Exception {
        log = new LoggingUtil(logger);
        log.start();

        // create the Joram Server
        joramAgent = new JoramAgent(log);
        joramAgent.startServer();

        containerName = SystemUtil.getContainerName();
        tcp = Integer.parseInt(SystemUtil.getJoramTCPPort());
        user = SystemUtil.getJoramUser();
        pwd = SystemUtil.getJoramPassword();
        host = SystemUtil.getHost();

        // create initialContext
        InitialContext context = new javax.naming.InitialContext(PropertyUtil
            .retrieveJNDIProperties());

        joramConnection = new JoramConnection(containerName, tcp, user, pwd,
            host, context);
        
        // Create an exception listener to log all asynchronious errors
        ExceptionListener exceptionListener = new ExceptionListener() {
            public void onException(JMSException exception) {
                log.error("Asynchronious JMS exception", exception);
                return;
            }
        };

        // Create a message listener to receive the messages from listerners
        // subscribers
        MessageListener messageListener = new MessageListener() {
            public void onMessage(Message aMessage) {
                processIncomingMessage(aMessage);
            }
        };

        try {
            joramConnection.start(exceptionListener, messageListener);
        } catch (Exception e) {
            throw new TransportException(e);
        }

        log.end();
    }

    /**
     * This method is called asynchronously when a Topic listener receives a
     * message. It is only responses messages of synchronious sends.
     * 
     * The received message is unserialiased into a
     * <code>MessageExchangeImpl</code> by using the <code>Serializer</code>,
     * and is pushed to the <code>Router</code>.
     * 
     * @param msg
     *            the jms message TODO processIncomingMessage :manage the
     *            reception of unknown message type. PB between the calculation
     *            of the next recipient and notion of service provider/consumer
     */
    protected void processIncomingMessage(Message msg) {
        log.start();
        if (msg instanceof ObjectMessage) {
            try {
                MessageExchangeImpl jbiMsg = serializer.jms2jbi(msg);
                router.pushResponse(jbiMsg);
            } catch (Exception e) {
                log.error(e.getMessage(), e);
            }
        } else {
            log
                .error("The received JMS message is not an ObjectMessage instance");
        }
        log.end();
    }

    /**
     * Close the JMS connection and stop the JORAM server. The
     * <code>Transporter</code> stay registered in the distributed JNDI
     * server.
     * 
     * @throws TransportException
     */
    @LifeCycle(on = LifeCycleType.STOP)
    public void stop() throws TransportException {
        log.start();
        try {
            joramConnection.stop();
            joramAgent.stopServer();
        } catch (Exception e) {
            String msg = "Can not stop the JMS connection.";
            log.error(msg, e);
            throw new TransportException(msg, e);
        }

        logger = null;
        log.end();
    }

    /**
     * Shutdown the <code>Transporter</code> and unregister it from the
     * distributed JNDI server. The embedded component's Topics must have been
     * removed from Joram and from the distributed Jndi server previously.
     * 
     * @throws TransportException
     */
    public void shutdown() throws TransportException {
        try {
            joramAgent.shutdown();
        } catch (Exception e) {
            String msg = "Can not shutdown correctly the Joram server.";
            log.error(msg, e);
            throw new TransportException(msg, e);
        }
    }

    /**
     * The <code>Transporter</code> creates a topic for the given component
     * and bind it to the JNDI server. It creates a subscriber to poll the
     * 'normal' messages and a listener to receiving the response messages when
     * synchronious send has been done.
     * 
     * @throws TransportException
     */
    public void addDestination(String componentName)
        throws TransportException {
        try {
            joramConnection.createTopicDevice(componentName);
        } catch (Exception e) {
            String msg = "Cannot create the topic device for the component '"
                + componentName + "'";
            log.error(msg, e);
            throw new TransportException(msg, e);
        }
    }

    /**
     * The <code>Transporter</code> removes the <code>Sesssions</code> bound
     * to the component related Topic. Then it deletes the Topic and unbind it
     * from the JNDI server.
     */
    public void removeDestination(String componentName)
        throws TransportException {
        try {
            joramConnection.deleteTopicDevice(componentName);
        } catch (Exception e) {
            String msg = "Cannot remove the topic for the component '"
                + componentName + "'";
            log.error(msg, e);
            throw new TransportException(msg, e);
        }
    }

    /**
     * The <code>Transporter</code> polls for a received message on the
     * component related topic.
     */
    public MessageExchangeImpl poll(String componentName, long timeoutMS,
        boolean block) throws TransportException {
        MessageExchangeImpl jbiMsg = null;

        try {
            Message receivedMessage = joramConnection.receiveForDestination(
                componentName, timeoutMS, block);

            // if receivedMessage is null, it means that the component has just
            // been uninstalled
            if (receivedMessage != null)
                jbiMsg = serializer.jms2jbi(receivedMessage);
        } catch (JMSException e) {
            StringBuilder msgBuilder = new StringBuilder();
            msgBuilder.append("Cannot retrieve message for the component '")
                .append(componentName).append("'");
            log.error(msgBuilder.toString(), e);
            throw new TransportException(msgBuilder.toString(), e);
        } catch (JBIException e) {
            StringBuilder msgBuilder = new StringBuilder();
            msgBuilder.append(
                "Cannot unserialize the retrieved message for the component '")
                .append(componentName).append("'");
            log.error(msgBuilder.toString(), e);
            throw new TransportException(msgBuilder.toString(), e);
        } catch (InterruptedException e) {
            StringBuilder msgBuilder = new StringBuilder();
            msgBuilder.append(
                "The thread trying to retrieve a message for the component '")
                .append(componentName).append("' has been interrrupted");
            log.error(msgBuilder.toString(), e);
            throw new TransportException(msgBuilder.toString(), e);
        }

        return jbiMsg;
    }
}