/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: Transporter.java,v 0.3 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.transport;

import org.objectweb.petals.PetalsLifeCycle;
import org.objectweb.petals.jbi.messaging.MessageExchangeImpl;

/**
 * The <code>Transporter</code> is used to send a <code>MessageExchange</code>
 * from a Petals container to another one. <br>
 * The local <code>Router</code> calls
 * <code>send(exchange,distantContainer)</code> on this
 * <code>Transporter</code> to send an exchange. <br>
 * This <code>Transporter</code> sends the <code>MessageExchange</code> to
 * the distant <code>Transporter</code>. <br>
 * When the distant <code>Transporter</code> receives the
 * <code>MessageExchange</code>, it has to call
 * <code>receive(exchange)</code> on its local <code>Router</code>. <br>
 * The localization of the distant container is in charge of the
 * <code>Transporter</code>. <br>
 * <br>
 * <em>Note that the destination container can be the same than this one.</em>
 * 
 * @version $Rev: 477 $ $Date: 2006-05-29 17:18:07 +0200 (lun., 29 mai 2006) $
 * @since Petals 1.0
 * @author alouis
 * @author rnaudin
 */
public interface Transporter extends PetalsLifeCycle{

    public final String PROPERTY_NOPERSISTANCE="org.objectweb.petals.transport.nonpersistent";
    
    /**
     * Send the specified exchange to the specified container.
     * If the "org.objectweb.petals.transport.nonpersistent" exchange
     * property is set to true, the message is not persisted by
     * the Transport layer, but it can be lost in case of problem.
     * @param exchange
     *            the exchange to send. <b>MUST</b> be non null
     * @param componentName
     *            the name of the distant container. <b>MUST</b> be non null
     * @param containerName
     *            the name of the distant container. <b>MUST</b> be non null
     *            and non empty
     * @param timeToLive
     *            duration of the message life (in ms). 0 for no timeout
     * @throws TransportException
     *             If the exchange is invalid, or can not be serialized. If the
     *             distant container can not be found.
     */
    void send(MessageExchangeImpl exchange, String componentName, 
        String containerName, long timeToLive) throws TransportException;

    /**
     * Inform the Transporter that the specified destination has been installed.
     * The transporter may need to initialise a device to handle the component
     * communications.
     *  
     * @param componentName
     *              The unique name of the installed component.
     *              <b>MUST</b> be non null
     * 
     * @throws TransportException
     */
    
    void addDestination(String componentName) throws TransportException;
    
    /**
     * Inform the transporter that the specified destination has been uninstalled.
     * The transporter may need to cleanup the device that handles the component
     * communications.
     *  
     * @param componentName
     *              The unique name of the uninstalled component.
     *              <b>MUST</b> be non null
     * 
     * @throws TransportException
     */

    void removeDestination(String componentName) throws TransportException;
    
    /**
     * Poll a received message for the given component
     * 
     * @param componentName
     *              The unique name of the component; s<b>MUST</b> be non null
     * @param timeoutMS
     *              The timeout in milli-seconds, specifies the poll duration
     * @param block
     *              If <code>True</code>, the poll is blocking until a response 
     *                            
     * @return The received message exchange
     */
    MessageExchangeImpl poll(String componentName, long timeoutMS, boolean block) throws TransportException;
}
