/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: MonitorBean.java msauvage $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.tools.webadmin.controller;

import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import javax.management.MBeanServerConnection;

import org.objectweb.petals.tools.webadmin.Object.ExchangesList;
import org.objectweb.petals.tools.webadmin.Object.Message;
import org.objectweb.petals.tools.webadmin.Object.MessageExchange;
import org.objectweb.petals.tools.webadmin.Object.ServerConfiguration;
import org.objectweb.petals.tools.webadmin.service.JMXService;
import org.objectweb.petals.tools.webadmin.util.JBIEntity;

public class MonitorController {

	/**
	 * host of the server
	 */
	private String host;

	/**
	 * JMX port of the server
	 */
	private String jmxPort;

	/**
	 * jmx user login for the server
	 */
	private String jmxLogin;

	/**
	 * jmx user password for the server
	 */
	private String jmxPassword;

	private final Lock lock = new ReentrantLock();

	/**
	 * Returns all exchanges on a server as ExchangesList object
	 * 
	 * @param connection
	 *            a MBeanServerConnection to the server
	 * @param sinceTime
	 *            timestamp of search beginning
	 * @return list of MessageExchange objects
	 * @throws Exception
	 */
	public ExchangesList getExchanges(MBeanServerConnection connection,
			long sinceTime) throws Exception {
		try {
			ExchangesList exchList = null;
			lock.lock();
			try {
				exchList = JMXService.getExchanges(connection, sinceTime);
			} finally {
				lock.unlock();
			}
			return exchList;
		} catch (Exception e) {
			throw new Exception(e);
		}
	}

	/**
	 * Returns exchanges on a server for a given component
	 * 
	 * @param connection
	 *            a MBeanServerConnection to the server
	 * @param sinceTime
	 *            timestamp of search beginning
	 * @param component
	 *            component name
	 * @return list of MessageExchange objects
	 * @throws Exception
	 */
	public ExchangesList getExchanges(MBeanServerConnection connection,
			long sinceTime, String component) throws Exception {
		try {
			// get the whole list
			ExchangesList completeList = this.getExchanges(connection,
					sinceTime);
			// creation of an empty list
			ExchangesList componentExchanges = new ExchangesList();
			for (MessageExchange currentExch : completeList) {
				// indicates if exchange concerns component asked
				boolean concernComponent = false;
				for (Message currentMsg : currentExch) {
					// component is found in a message of the exchange
					// -> mark as concerned
					if (currentMsg.getComponent().equals(component)) {
						concernComponent = true;
					}
				}
				// current exchange has at least one message for the asked
				// component
				// -> add to the list
				if (concernComponent) {
					componentExchanges.add(currentExch);
				}
			}
			return componentExchanges;
		} catch (Exception e) {
			throw new Exception(e);
		}
	}

	/**
	 * adds the ExchangesList of a server in a Map. Uses the serverName as key
	 * for the map. If an ExchangesList already exists for this server, adds the
	 * new Exchanges in the ExchangesList. If not, create a new entry.
	 * 
	 * @param storage_map
	 *            the Map in wich Exchanges must be added
	 * @param serverName
	 *            name of the server containing exchanges
	 * @param exchanges
	 *            new exchanges to add
	 */
	public void setExchanges(Map<String, ExchangesList> storage_map,
			String serverName, ExchangesList exchanges) {
		lock.lock();
		try {
			if (storage_map.get(serverName) != null) {
				ExchangesList serverExchanges = storage_map.get(serverName);
				for (MessageExchange currentNewExchange : exchanges) {
					// adding the exchange only if doesn't already exists in
					// list
					if (!serverExchanges
							.hasExchange(currentNewExchange.getId())) {
						serverExchanges.add(currentNewExchange);
					} else {
						// exchange already exists in list -> adding messages
						MessageExchange existingExchange = serverExchanges
								.getExchange(currentNewExchange.getId());
						existingExchange.addAll(currentNewExchange);
					}
				}
			} else {
				storage_map.put(serverName, exchanges);
			}
		} finally {
			lock.unlock();
		}
	}

	/**
	 * Retrieve all servers actually started on the domain
	 * 
	 * @return list of server's configurations
	 * @throws Exception
	 */
	public List<ServerConfiguration> getStartedServers() throws Exception {
		try {
			ServerConfiguration serverConf = new ServerConfiguration(host,
					jmxPort, jmxLogin, jmxPassword);
			return JMXService.getStartedServers(serverConf);
		} catch (Exception e) {
			throw new Exception(e);
		}
	}

	public List<ServerConfiguration> getStartedServers(
			MBeanServerConnection connection) throws Exception {
		try {
			return JMXService.getStartedServers(connection);
		} catch (Exception e) {
			throw new Exception(e);
		}
	}

	/**
	 * Returns all servers of the domain, started and stopped
	 * 
	 * @return list of server's configurations
	 * @throws Exception
	 */
	public List<ServerConfiguration> getAllServers() throws Exception {
		try {
			ServerConfiguration serverConf = new ServerConfiguration(host,
					jmxPort, jmxLogin, jmxPassword);
			return JMXService.getAllServers(serverConf);
		} catch (Exception e) {
			throw new Exception(e);
		}
	}

	/**
	 * Extracts a server configuration from a servers list
	 * 
	 * @param servers
	 *            list of the servers configurations
	 * @param serverName
	 *            name of the server wanted
	 * @return ServerConfiguration, null if not found
	 */
	public ServerConfiguration getServerConf(List<ServerConfiguration> servers,
			String serverName) {
		boolean serverFound = false;
		ServerConfiguration serverConf = null;
		ListIterator<ServerConfiguration> serverLI = servers.listIterator();
		while (!serverFound && serverLI.hasNext()) {
			ServerConfiguration currentServer = serverLI.next();
			if (currentServer.getName().equals(serverName)) {
				serverFound = true;
				serverConf = currentServer;
			}
		}
		return serverConf;
	}

	/**
	 * list components of a specified type started on a server
	 * 
	 * @param serverConf
	 *            configuration of the server
	 * @param type
	 *            type of components BC for Binding Components, SE for Service
	 *            Engine, ALL for both
	 * @return Map of the component as <Name,Component>
	 * @throws Exception
	 */
	public Map<String, JBIEntity> listComponents(
			MBeanServerConnection connection, String type) throws Exception {
		try {
			return JMXService.listComponents(connection, type);
		} catch (Exception e) {
			throw new Exception(e);
		}
	}

	/**
	 * start or stop monitoring on a specific server
	 * 
	 * @param connection
	 *            a MBeanServerConnection on the server
	 * @param action
	 *            "start" to start monitoring, "stop" to stop monitoring
	 * @return
	 */
	public void monitor(MBeanServerConnection connection, String action)
			throws Exception {
		try {
			boolean boolAction = false;
			if (action.equals("start"))
				boolAction = true;
			JMXService.monitor(connection, boolAction);
		} catch (Exception e) {
			throw new Exception(e);
		}
	}

	/**
	 * start or stop monitoring on servers
	 * 
	 * @param startedServers
	 *            list of configurations of servers to monitor
	 * @param action
	 *            "start" to start monitoring, "stop" to stop monitoring
	 * @throws Exception
	 */
	public void monitorAllServers(List<ServerConfiguration> servers,
			String action) throws Exception {
		try {
			boolean boolAction = false;
			if (action.equals("start"))
				boolAction = true;
			if (servers != null && servers.size() > 0) {
				for (ServerConfiguration serverConf : servers) {
					JMXService.monitor(serverConf.getConnection(), boolAction);
				}
			}
		} catch (Exception e) {
			throw new Exception(e);
		}
	}

	/**
	 * Retrieves an attribute from the server
	 * 
	 * @param connection
	 *            a MBeanServerConnection to the server
	 * @param attributeName
	 *            name of the attribute to retrieve
	 * @return the value of the attribute
	 * @throws Exception
	 */
	public String getAttribute(MBeanServerConnection connection,
			String attributeName) throws Exception {
		try {
			return JMXService.getAttribute(connection, attributeName);
		} catch (Exception e) {
			throw new Exception(e);
		}
	}

	/**
	 * Returns number of messages in a server with timestamp corresponding to
	 * the second of a given timestamp
	 * 
	 * @param timestamp
	 *            timestamp of the second wanted
	 * @return number of messages
	 */
	public long getNbMsgByTime(ExchangesList exchanges, long timestamp) {
		lock.lock();
		try {
			return exchanges.getNbMsgByTime(timestamp);
		} finally {
			lock.unlock();
		}
	}

	/**
	 * Returns the number of messages concerning a given component in a list of
	 * Exchanges
	 * 
	 * @param exchanges
	 *            the list of exchanges, as a ExchangesList object
	 * @param componentName
	 *            the name of the component requested
	 * @return a number of messages as a long
	 */
	public long getNbMsgByComponent(ExchangesList exchanges,
			String componentName) {
		lock.lock();
		try {
			return exchanges.getNbMsgByComponent(componentName);
		} finally {
			lock.unlock();
		}
	}

	/**
	 * Returns a list of exchanges containing messages of a global ExchangesList
	 * concerning a specific component
	 * 
	 * @param exchanges
	 *            list of all exchanges
	 * @param componentName
	 *            name of the component searched
	 * @return a list of MessageExchanges
	 */
	public ExchangesList getMsgByComponent(ExchangesList exchanges,
			String componentName) {
		lock.lock();
		try {
			return exchanges.getByComponent(componentName);
		} finally {
			lock.unlock();
		}
	}

	/**
	 * Returns the host of the server
	 * 
	 * @return a string representation of the server's host (ip, dns name, etc.)
	 *         as setted
	 */
	public String getHost() {
		return host;
	}

	/**
	 * Sets the host of the server
	 * 
	 * @param host
	 *            a String that is to be the server's host
	 */
	public void setHost(String host) {
		this.host = host;
	}

	/**
	 * Returns the JMX port
	 * 
	 * @return a string representation of the port
	 */
	public String getJmxPort() {
		return jmxPort;
	}

	/**
	 * Sets the JMX port
	 * 
	 * @param jmxPort
	 *            a String that is to be the port
	 */
	public void setJmxPort(String jmxPort) {
		this.jmxPort = jmxPort;
	}

	/**
	 * Returns the JMX login name
	 * 
	 * @return the JMX login name
	 */
	public String getJmxLogin() {
		return jmxLogin;
	}

	/**
	 * Sets the JMX login name
	 * 
	 * @param jmxLogin
	 *            a String that is to be the JMX login name
	 */
	public void setJmxLogin(String jmxLogin) {
		this.jmxLogin = jmxLogin;
	}

	/**
	 * Returns the JMX user password
	 * 
	 * @return the jmx password
	 */
	public String getJmxPassword() {
		return jmxPassword;
	}

	/**
	 * Sets the JMX user password
	 * 
	 * @param jmxPassword
	 *            a String that is to be password for jmx access
	 */
	public void setJmxPassword(String jmxPassword) {
		this.jmxPassword = jmxPassword;
	}
}
