/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: JMXService.java msauvage $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.tools.webadmin.service;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.management.Attribute;
import javax.management.AttributeNotFoundException;
import javax.management.MBeanServerConnection;
import javax.management.ObjectName;
import javax.management.remote.JMXConnector;

import org.objectweb.petals.tools.webadmin.Object.ExchangesList;
import org.objectweb.petals.tools.webadmin.Object.ServerConfiguration;
import org.objectweb.petals.tools.webadmin.util.JBIEntity;
import org.objectweb.petals.tools.webadmin.util.JBIJMXConnectorUtil;
import org.objectweb.petals.tools.webadmin.util.JMXUtils;

public class JMXService {

	/**
	 * indicates if a server is accessible via JMX or not
	 * 
	 * @param serverConf
	 *            the server's configuration
	 * @return true if server is accessible, false if not accessible
	 * @throws IOException
	 */
	public static boolean serverReachable(ServerConfiguration serverConf)
			throws IOException {
		try {
			if (JMXUtils.retrieveConnector(serverConf) != null)
				return true;
			else
				return false;
		} catch (IOException ioe) {
			return false;
		}
	}

	/**
	 * Lists the deployed JBI components of a specified type
	 * 
	 * @param connection
	 *            a MBeanServerConnection to the server
	 * @param type
	 *            type of component : BC for Binding Components, SE for Service
	 *            Engine, ALL for both BC and SE
	 * @return a Map of JBIEntity components as <name,component>
	 * @throws Exception
	 */
	public static Map<String, JBIEntity> listComponents(
			MBeanServerConnection connection, String type) throws Exception {
		Map<String, JBIEntity> components = new HashMap<String, JBIEntity>();
		try {
			Object result;
			if (type.equals("BC") || type.equals("ALL")) {
				result = connection.getAttribute(JBIJMXConnectorUtil
						.getAdminServiceMBeanName(connection),
						"BindingComponents");
				if (result instanceof ObjectName[]) {
					ObjectName[] engines = (ObjectName[]) result;
					for (ObjectName objectName : engines) {
						JBIEntity component = new JBIEntity();
						String objState = (String) connection.getAttribute(
								objectName, "CurrentState");
						String name = objectName.getKeyProperty("name");
						component.setName(name);
						component.setState(objState);
						component.setObjectName(objectName);
						// TODO find a way to get the description of a component
						component.setDescription("");
						component.setType(JBIEntity.BINDING_COMPONENT);
						components.put(component.getName(), component);
					}
				}
			}
			if (type.equals("SE") || type.equals("ALL")) {
				result = connection.getAttribute(JBIJMXConnectorUtil
						.getAdminServiceMBeanName(connection),
						"EngineComponents");
				if (result instanceof ObjectName[]) {
					ObjectName[] engines = (ObjectName[]) result;
					for (ObjectName objectName : engines) {
						JBIEntity component = new JBIEntity();
						String objState = (String) connection.getAttribute(
								objectName, "CurrentState");
						String name = objectName.getKeyProperty("name");
						component.setName(name);
						component.setState(objState);
						component.setObjectName(objectName);
						// TODO find a way to get the description of a component
						component.setDescription("");
						component.setType(JBIEntity.SERVICE_ENGINE);
						components.put(component.getName(), component);
					}
				}
			}
		} catch (Exception e) {
			throw new Exception(e);
		}
		return components;
	}

	/**
	 * Return all components started on a server
	 * 
	 * @param connection
	 *            a MBeanServerConnection to the server
	 * @return
	 */
	public static Map<String, JBIEntity> listComponents(
			MBeanServerConnection connection) {
		try {
			return listComponents(connection, "ALL");
		} catch (Exception e) {
			e.printStackTrace(System.out);
			return null;
		}
	}

	/**
	 * Returns the names of the components started on a server
	 * 
	 * @param connection
	 *            a MBeanServerConnection to the server
	 * @param type
	 *            type of component : BC for Binding Components, SE for Service
	 *            Engine, ALL for both BC and SE
	 * @return
	 */
	public static List<String> getComponentsName(
			MBeanServerConnection connection, String type) throws Exception {
		List<String> components = new ArrayList<String>();
		try {
			Object result;
			if (type.equals("BC") || type.equals("ALL")) {
				result = connection.getAttribute(JBIJMXConnectorUtil
						.getAdminServiceMBeanName(connection),
						"BindingComponents");
				if (result instanceof ObjectName[]) {
					ObjectName[] engines = (ObjectName[]) result;
					for (ObjectName objectName : engines) {
						components.add(objectName.getKeyProperty("name"));
					}
				}
			}
			if (type.equals("SE") || type.equals("ALL")) {
				result = connection.getAttribute(JBIJMXConnectorUtil
						.getAdminServiceMBeanName(connection),
						"EngineComponents");
				if (result instanceof ObjectName[]) {
					ObjectName[] engines = (ObjectName[]) result;
					for (ObjectName objectName : engines) {
						components.add(objectName.getKeyProperty("name"));
					}
				}
			}
		} catch (Exception e) {
			throw new Exception(e);
		}
		return components;
	}

	/**
	 * Retrieve all servers actually started on a domain
	 * 
	 * @param serverConf
	 *            the ServerConfiguration of a server known on the domain
	 * @return list of server's configurations
	 * @throws Exception
	 */
	public static List<ServerConfiguration> getStartedServers(
			ServerConfiguration serverConf) throws Exception {
		// list of the servers name to return
		List<ServerConfiguration> serversList = getAllServers(serverConf);
		//list of server's index in servers list to remove
		List<Integer> toRemove = new ArrayList<Integer>();
		// search for servers to remove -> adding them index to remove list
		for (ServerConfiguration currentServer : serversList) {
			if (currentServer.getState().equals("stopped")) {
				toRemove.add(new Integer(serversList.indexOf(currentServer)));
			}
		}
		//removing servers
		for (Integer serverIndex : toRemove)
		{
			serversList.remove(serverIndex.intValue());
		}
		return serversList;
	}

	/**
	 * Retrieve all servers actually started on a domain, passing through a
	 * given server connection
	 * 
	 * @param serverConf
	 *            the ServerConfiguration of a server known on the domain
	 * @return list of server's configurations
	 * @throws Exception
	 */
	public static List<ServerConfiguration> getStartedServers(
			MBeanServerConnection connection) throws Exception {
		// list of the servers name to return
		List<ServerConfiguration> serversList = getAllServers(connection);
		//list of server's index in servers list to remove
		List<Integer> toRemove = new ArrayList<Integer>();
		// search for servers to remove -> adding them index to remove list
		for (ServerConfiguration currentServer : serversList) {
			if (currentServer.getState().equals("stopped")) {
				toRemove.add(new Integer(serversList.indexOf(currentServer)));
			}
		}
		//removing servers
		for (Integer serverIndex : toRemove)
		{
			serversList.remove(serverIndex.intValue());
		}
		return serversList;
	}

	/**
	 * Returns all servers of the domain, started and stopped
	 * 
	 * @param serverConf
	 *            the ServerConfiguration of a server known on the domain
	 * @return list of server's configurations
	 * @throws Exception
	 */
	@SuppressWarnings("unchecked")
	public static List<ServerConfiguration> getAllServers(
			ServerConfiguration serverConf) throws Exception {
		List<ServerConfiguration> serversList = new ArrayList<ServerConfiguration>();
		try {
			JMXConnector connector = JMXUtils.retrieveConnector(serverConf);
			MBeanServerConnection connection = connector
					.getMBeanServerConnection();
			ObjectName petalsAdminName = new ObjectName(
					"Petals:name=PetalsAdmin,type=service");
			Collection<Map<String, String>> serversConfigurations = (Collection<Map<String, String>>) connection
					.getAttribute(petalsAdminName, "AllServersConfiguration");
			for (Map<String, String> currentServer : serversConfigurations) {
				ServerConfiguration currentServerConf = new ServerConfiguration(
						currentServer.get("name"), currentServer.get("host"),
						currentServer.get("jmxPort"), currentServer
								.get("jmxLogin"), currentServer
								.get("jmxPassword"), currentServer.get("state"));
				if (currentServerConf.getState().equals("started")) {
					// creation of a connection for the server
					MBeanServerConnection currentServerConnection = JMXUtils
							.retrieveConnector(currentServerConf)
							.getMBeanServerConnection();
					currentServerConf.setConnection(currentServerConnection);
				}
				serversList.add(currentServerConf);
			}

		} catch (Exception e) {
			throw new Exception(e);
		}
		return serversList;
	}

	/**
	 * Returns all servers of the domain, started and stopped, passing through a
	 * given server connection
	 * 
	 * @param serverConf
	 *            the ServerConfiguration of a server known on the domain
	 * @return list of server's configurations
	 * @throws Exception
	 */
	@SuppressWarnings("unchecked")
	public static List<ServerConfiguration> getAllServers(
			MBeanServerConnection connection) throws Exception {
		List<ServerConfiguration> serversList = new ArrayList<ServerConfiguration>();
		try {
			ObjectName petalsAdminName = new ObjectName(
					"Petals:name=PetalsAdmin,type=service");
			Collection<Map<String, String>> serversConfigurations = (Collection<Map<String, String>>) connection
					.getAttribute(petalsAdminName, "AllServersConfiguration");
			for (Map<String, String> currentServer : serversConfigurations) {
				ServerConfiguration currentServerConf = new ServerConfiguration(
						currentServer.get("name"), currentServer.get("host"),
						currentServer.get("jmxPort"), currentServer
								.get("jmxLogin"), currentServer
								.get("jmxPassword"), currentServer.get("state"));
				if (currentServerConf.getState().equals("started")) {
					// creation of a connection for the server
					MBeanServerConnection currentServerConnection = JMXUtils
							.retrieveConnector(currentServerConf)
							.getMBeanServerConnection();
					currentServerConf.setConnection(currentServerConnection);
				}
				serversList.add(currentServerConf);
			}
		} catch (Exception e) {
			throw new Exception(e);
		}
		return serversList;
	}

	/**
	 * Returns a server's name, based on his jmx informations
	 * 
	 * @param serverConf
	 *            the configuration of the server
	 * @return the name of the server, null if not found
	 */
	public static String getServerName(ServerConfiguration serverConf) {
		try {
			List<ServerConfiguration> serversList = getAllServers(serverConf);
			for (ServerConfiguration currentServer : serversList) {
				if (currentServer.getHost().equals(serverConf.getHost())
						&& currentServer.getJmxPort().equals(
								serverConf.getJmxPort())) {
					return currentServer.getName();
				}
			}
			return null;
		} catch (Exception e) {
			return null;
		}
	}

	/**
	 * Returns all exchanges from a given server since a given time
	 * 
	 * @param connection
	 *            a MBeanServerConnection to the server
	 * @param sinceTime
	 *            timestamp of search beginning
	 * @return list of MessageExchange objects
	 * @throws Exception
	 */
	@SuppressWarnings("unchecked")
	public static ExchangesList getExchanges(MBeanServerConnection connection,
			long sinceTime) throws Exception {
		try {
			ObjectName petalsAdminName = new ObjectName(
					"Petals:name=PetalsAdmin,type=service");
			return new ExchangesList(
					(List<List<Map<String, Object>>>) connection.invoke(
							petalsAdminName, "bulkData",
							new Object[] { sinceTime },
							new String[] { Long.TYPE.getCanonicalName() }));
		} catch (Exception e) {
			throw new Exception(e);
		}
	}

	/**
	 * start or stop monitoring on a specific server
	 * 
	 * @param connection
	 *            a MBeanServerConnection to the server
	 * @param action :
	 *            "start" to start monitoring, "stop" to stop monitoring
	 * @return
	 */
	public static void monitor(MBeanServerConnection connection, boolean action)
			throws Exception, AttributeNotFoundException {
		try {
			ObjectName petalsAdminName = new ObjectName(
					"Petals:name=PetalsAdmin,type=service");
			Attribute attAction = new Attribute("Monitoring", action);
			connection.setAttribute(petalsAdminName, attAction);
		} catch (AttributeNotFoundException anf) {
			throw new AttributeNotFoundException();
		} catch (Exception e) {
			throw new Exception(e);
		}
	}

	/**
	 * Retrieves an attribute from the server
	 * 
	 * @param connection
	 *            a MBeanServerConnection to the server
	 * @param attributeName
	 *            name of the attribute to retrieve
	 * @return the value of the attribute
	 * @throws Exception
	 * @throws AttributeNotFoundException
	 */
	public static String getAttribute(MBeanServerConnection connection,
			String attributeName) throws Exception, AttributeNotFoundException {
		try {
			ObjectName petalsAdminName = new ObjectName(
					"Petals:name=PetalsAdmin,type=service");
			return connection.getAttribute(petalsAdminName, attributeName)
					.toString();
		} catch (AttributeNotFoundException anf) {
			return "not found";
		} catch (Exception e) {
			throw new Exception(e);
		}
	}

}
